# coding=utf-8
__author__ = "Dimitris Karkalousos"

import contextlib
from typing import List, Optional, Sequence, Tuple, Union

import numba as nb
import numpy as np
import torch


@contextlib.contextmanager
def temp_seed(rng: np.random, seed: Optional[Union[int, Tuple[int, ...]]]):
    """Temporarily set the seed of a numpy random number generator.

    Parameters
    ----------
    rng : np.random.Generator
        The numpy random number generator to modify.
    seed : Optional[Union[int, Tuple[int, ...]]], optional
        The seed to set, by default None.
    """
    if seed is None:
        try:
            yield
        finally:
            pass
    else:
        state = rng.get_state()
        rng.seed(seed)
        try:
            yield
        finally:
            rng.set_state(state)


class MaskFunc:
    """MaskFunc is an abstract base class for creating sub-sampling masks.

    This class is used to create a mask for MRI data that can be used to randomly under-sample the k-space data. The
    mask is created by retaining a specified fraction of the low-frequency columns and setting the rest to zero. The
    fraction of low-frequency columns to retain and the amount of under-sampling can be specified at initialization.

    Examples
    --------
    >>> from atommic.collections.common.data.subsample import MaskFunc
    >>> mask_func = MaskFunc(center_fractions=[0.08, 0.04], accelerations=[4, 8])
    >>> mask_func.choose_acceleration()
    (0.08, 4)
    """

    def __init__(self, center_fractions: Sequence[float], accelerations: Sequence[int]):
        """Inits :class:`MaskFunc`.

        Parameters
        ----------
        center_fractions : Sequence[float]
            Fraction of low-frequency columns to be retained. If multiple values are provided, then one of these
            numbers is chosen uniformly each time. For 2D setting this value corresponds to setting the
            Full-Width-Half-Maximum.
        accelerations : Sequence[int]
            Amount of under-sampling. This should have the same length as center_fractions. If multiple values are
            provided, then one of these is chosen uniformly each time.
        """
        super().__init__()
        self.center_fractions = center_fractions
        self.accelerations = accelerations
        self.rng = np.random.RandomState()

    def __call__(
        self,
        shape: Sequence[int],
        seed: Optional[Union[int, Tuple[int, ...]]] = None,
        partial_fourier_percentage: Optional[float] = 0.0,
        **kwargs,
    ) -> Tuple[torch.Tensor, int]:
        """Calls :class:`MaskFunc`.

        Parameters
        ----------
        shape : Sequence[int]
            The shape of the mask to be created. The shape should have at least 3 dimensions. Same as the shape of the
            input k-space data.
        seed : int or tuple of ints, optional
            Seed for the random number generator. Default is ``None``.
        partial_fourier_percentage : float, optional
            Percentage of the low-frequency columns to be retained. Default is ``0.0``.
        """
        raise NotImplementedError

    def choose_acceleration(self) -> Tuple[float, int]:
        """Chooses an acceleration factor and center fractions from a list of multiple values.

        Returns
        -------
        Tuple[float, int]
            A tuple of the center fraction and the acceleration factor.
        """
        choice = self.rng.randint(0, len(self.accelerations))
        center_fraction = self.center_fractions[choice]
        acceleration = self.accelerations[choice]
        return center_fraction, acceleration


class Equispaced1DMaskFunc(MaskFunc):
    r"""Equispaced1DMaskFunc creates a sub-sampling mask of a given shape.

    The mask selects a subset of columns from the input k-space data. If the k-space data has N columns, the mask
    picks out:

        1. N_low_freqs = (N * center_fraction) columns in the center corresponding to low-frequencies.

        2. The other columns are selected with equal spacing at a proportion that reaches the desired acceleration
        rate taking into consideration the number of low frequencies. This ensures that the expected number of
        columns selected is equal to (N / acceleration).


    It is possible to use multiple center_fractions and accelerations, in which case one possible (center_fraction,
    acceleration) is chosen uniformly at random each time the Equispaced1DMaskFunc object is called.

    Note that this function may not give equispaced samples \
    (documented in https://github.com/facebookresearch/fastMRI/issues/54), which will require modifications to
    standard GRAPPA approaches. Nonetheless, this aspect of the function has been preserved to match the public
    multicoil data.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.data.subsample import Equispaced1DMaskFunc
    >>> mask_func = Equispaced1DMaskFunc(center_fractions=[0.08, 0.04], accelerations=[4, 8])
    >>> mask_func.choose_acceleration()
    (0.08, 4)
    >>> kspace = torch.randn(1, 1, 640, 368)
    >>> mask, acceleration = mask_func(kspace.shape)
    >>> mask.shape
    torch.Size([1, 1, 1, 368])
    >>> acceleration
    4
    """

    def __call__(
        self,
        shape: Sequence[int],
        seed: Optional[Union[int, Tuple[int, ...]]] = None,
        partial_fourier_percentage: Optional[float] = 0.0,
        **kwargs,
    ) -> Tuple[torch.Tensor, int]:
        """Calls :class:`Equispaced1DMaskFunc`.

        Parameters
        ----------
        shape : Sequence[int]
            The shape of the mask to be created. The shape should have at least 3 dimensions. Same as the shape of the
            input k-space data.
        seed : int or tuple of ints, optional
            Seed for the random number generator. Default is ``None``.
        partial_fourier_percentage : float, optional
            Percentage of the low-frequency columns to be retained. Default is ``0.0``.

        Returns
        -------
        Tuple[torch.Tensor, int]
            A tuple of the generated mask and the acceleration factor.

        Raises
        ------
        ValueError
            If the `shape` parameter has less than 3 dimensions.
        """
        if len(shape) < 3:
            raise ValueError("Shape should have 3 or more dimensions")

        with temp_seed(self.rng, seed):
            center_fraction, acceleration = self.choose_acceleration()
            num_cols = shape[-2]
            num_low_freqs = int(round(num_cols * center_fraction))

            # create the mask
            mask = np.zeros(num_cols, dtype=np.float32)
            pad = torch.div((num_cols - num_low_freqs + 1), 2, rounding_mode="trunc").item()
            mask[pad : pad + num_low_freqs] = True

            # determine acceleration rate by adjusting for the number of low frequencies
            adjusted_accel = (acceleration * (num_low_freqs - num_cols)) / (num_low_freqs * acceleration - num_cols)
            offset = self.rng.randint(0, round(adjusted_accel))

            accel_samples = np.arange(offset, num_cols - 1, adjusted_accel)
            accel_samples = np.around(accel_samples).astype(np.uint)
            mask[accel_samples] = True

            # reshape the mask
            mask_shape = [1 for _ in shape]
            mask_shape[-2] = num_cols
            mask = torch.from_numpy(mask.reshape(*mask_shape).astype(np.float32))

            if partial_fourier_percentage != 0:
                mask[:, : int(np.round(mask.shape[1] * partial_fourier_percentage))] = 0.0

        return mask, acceleration


class Equispaced2DMaskFunc(MaskFunc):
    """Same as Equispaced1DMaskFunc, but for 2D k-space data.

    .. note::
        See ..class::`atommic.collections.common.data.subsample.Equispaced1DMaskFunc` for more details.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.data.subsample import Equispaced2DMaskFunc
    >>> mask_func = Equispaced2DMaskFunc(center_fractions=[0.08, 0.04], accelerations=[4, 8])
    >>> mask_func.choose_acceleration()
    (0.08, 4)
    >>> kspace = torch.randn(1, 1, 640, 368)
    >>> mask, acceleration = mask_func(kspace.shape)
    >>> mask.shape
    torch.Size([1, 1, 640, 368])
    >>> acceleration
    4
    """

    def __call__(
        self,
        shape: Sequence[int],
        seed: Optional[Union[int, Tuple[int, ...]]] = None,
        partial_fourier_percentage: Optional[float] = 0.0,
        **kwargs,
    ) -> Tuple[torch.Tensor, int]:
        """Calls :class:`Equispaced2DMaskFunc`.

        Parameters
        ----------
        shape : Sequence[int]
            The shape of the mask to be created. The shape should have at least 3 dimensions. Same as the shape of the
            input k-space data.
        seed : int or tuple of ints, optional
            Seed for the random number generator. Default is ``None``.
        partial_fourier_percentage : float, optional
            Percentage of the low-frequency columns to be retained. Default is ``0.0``.

        Returns
        -------
        Tuple[torch.Tensor, int]
            A tuple containing the mask and the acceleration factor.
            The mask is a `torch.Tensor` of the same shape as the input shape.
            The acceleration factor is an `int` that represents the number of samples taken in the k-space.

        Raises
        ------
        ValueError
            If the `shape` parameter has less than 3 dimensions.
        """
        if len(shape) < 3:
            raise ValueError("Shape should have 3 or more dimensions")

        with temp_seed(self.rng, seed):
            center_fraction, acceleration = self.choose_acceleration()

            acceleration = int(acceleration / 2)
            center_fraction = center_fraction / 2

            num_cols = shape[-2]
            num_low_freqs = int(round(num_cols * center_fraction))

            num_rows = shape[-3]
            num_high_freqs = int(round(num_rows * center_fraction))

            # create the mask
            mask = np.zeros([num_rows, num_cols], dtype=np.float32)

            pad_cols = torch.div((num_cols - num_low_freqs + 1), 2, rounding_mode="trunc").item()
            pad_rows = torch.div((num_rows - num_high_freqs + 1), 2, rounding_mode="trunc").item()
            mask[pad_rows : pad_rows + num_high_freqs, pad_cols : pad_cols + num_low_freqs] = True

            for i in np.arange(0, num_rows, acceleration):
                for j in np.arange(0, num_cols, acceleration):
                    mask[int(i), int(j)] = True

            # reshape the mask
            mask_shape = [1 for _ in shape]
            mask_shape[-2] = num_cols
            mask_shape[-3] = num_rows
            mask = torch.from_numpy(mask.reshape(*mask_shape).astype(np.float32))

            if partial_fourier_percentage != 0:
                mask[:, : int(np.round(mask.shape[1] * partial_fourier_percentage))] = 0.0

        return mask, acceleration * 2


class Gaussian1DMaskFunc(MaskFunc):
    """Same as Gaussian2DMaskFunc, but for 1D k-space data.

    .. note::
        See ..class::`atommic.collections.common.data.subsample.Gaussian2DMaskFunc` for more details.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.data.subsample import Gaussian1DMaskFunc
    >>> mask_func = Gaussian1DMaskFunc(center_fractions=[0.7, 0.7], accelerations=[4, 8])
    >>> mask_func.choose_acceleration()
    (0.7, 4)
    >>> kspace = torch.randn(1, 1, 640, 368)
    >>> mask, acceleration = mask_func(kspace.shape)
    >>> mask.shape
    torch.Size([1, 1, 1, 368])
    >>> acceleration
    4
    """

    def __call__(
        self,
        shape: Union[Sequence[int], np.ndarray],
        seed: Optional[Union[int, Tuple[int, ...]]] = None,
        partial_fourier_percentage: Optional[float] = 0.0,
        center_scale: Optional[float] = 0.02,
        **kwargs,
    ) -> Tuple[torch.Tensor, int]:
        """Calls :class:`Gaussian1DMaskFunc`.

        Parameters
        ----------
        shape : Sequence[int]
            The shape of the mask to be created. The shape should have at least 3 dimensions. Same as the shape of the
            input k-space data.
        seed : int or tuple of ints, optional
            Seed for the random number generator. Default is ``None``.
        partial_fourier_percentage : float, optional
            Percentage of the low-frequency columns to be retained. Default is ``0.0``.
        center_scale : float, optional
            For autocalibration purposes, data points near the k-space center will be fully sampled within an ellipse
            of which the half-axes will be set to the given `center_scale` percentage of the fully sampled region.
            Default is ``0.02``.

        Returns
        -------
        Tuple[torch.Tensor, int]
            A tuple of the generated mask and the acceleration factor.
        """
        with temp_seed(self.rng, seed):
            dims = [1 for _ in shape]
            self.shape = tuple(shape[-3:-1])
            self.shape = (self.shape[1], self.shape[0])
            dims[-2] = self.shape[-2]

            full_width_half_maximum, acceleration = self.choose_acceleration()

            self.full_width_half_maximum = full_width_half_maximum
            self.acceleration = acceleration
            self.center_scale = center_scale

            mask = self.gaussian_kspace()
            mask[tuple(self.gaussian_coordinates())] = 1.0

            mask = np.fft.ifftshift(np.fft.ifftshift(np.fft.ifftshift(mask, axes=0), axes=0), axes=(0, 1))

            if partial_fourier_percentage != 0:
                mask[:, : int(np.round(mask.shape[1] * partial_fourier_percentage))] = 0.0

            mask = torch.from_numpy(np.transpose(mask, (1, 0))[0].reshape(*dims).astype(np.float32))

        return mask, acceleration

    def gaussian_kspace(self) -> np.ndarray:
        """Creates a Gaussian sampled k-space center."""
        scaled = int(self.shape[0] * self.center_scale)
        center = np.ones((scaled, self.shape[1]))
        top_scaled = torch.div((self.shape[0] - scaled), 2, rounding_mode="trunc").item()
        bottom_scaled = self.shape[0] - scaled - top_scaled
        top = np.zeros((top_scaled, self.shape[1]))
        btm = np.zeros((bottom_scaled, self.shape[1]))
        return np.concatenate((top, center, btm))

    def gaussian_coordinates(self) -> Tuple[np.ndarray, np.ndarray]:
        r"""Returns Gaussian sampled k-space coordinates.

        Returns
        -------
        xsamples : np.ndarray
            A 1D numpy array of x-coordinates.
        ysamples : np.ndarray
            A 1D numpy array of y-coordinates.

        Notes
        -----
        The number of samples taken is determined by `n_sample` which is calculated as
        `self.shape[0] / self.acceleration`. The selection of the samples is based on the probabilities calculated
        from `gaussian_kernel`.
        """
        n_sample = int(self.shape[0] / self.acceleration)
        idxs = np.random.choice(range(self.shape[0]), size=n_sample, replace=False, p=self.gaussian_kernel())
        xsamples = np.concatenate([np.tile(i, self.shape[1]) for i in idxs])
        ysamples = np.concatenate([range(self.shape[1]) for _ in idxs])
        return xsamples, ysamples

    def gaussian_kernel(self) -> np.ndarray:
        r"""Creates a Gaussian sampled k-space kernel.

        .. note::
            The function calculates the Gaussian kernel by computing the sum of the exponential of the squared \
            x-values divided by 2 times the square of the standard deviation. The standard deviation is calculated \
            from the full width at half maximum (FWHM) of the Gaussian curve and is defined as the FWHM divided by \
            the square root of 8 times the natural logarithm of 2. The FWHM and the kern_len are obtained from the \
            `full_width_half_maximum` and `shape` attributes of the class respectively.

        Returns
        -------
        ndarray
            The Gaussian kernel.
        """
        kernel = 1
        for kern_len in self.shape:
            sigma = self.full_width_half_maximum / np.sqrt(8 * np.log(2))
            x = np.linspace(-1.0, 1.0, kern_len)
            g = np.exp(-(x**2 / (2 * sigma**2)))  # noqa: F841
            kernel = g
            break
        kernel = kernel / kernel.sum()  # type: ignore
        return kernel


class Gaussian2DMaskFunc(MaskFunc):
    """Creates a 2D sub-sampling mask of a given shape.

    The sub-sampling mask is generated in k-space, with data points near the k-space center being fully sampled within
    an ellipse. The half-axes of the ellipse are set to the `center_scale` percentage of the fully sampled region. The
    remaining points are sampled according to a Gaussian distribution.

    The center fractions act as Full-Width at Half-Maximum (FWHM) values.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.common.data.subsample import Gaussian2DMaskFunc
    >>> mask_func = Gaussian2DMaskFunc(center_fractions=[0.7, 0.7], accelerations=[4, 8])
    >>> mask_func.choose_acceleration()
    (0.7, 4)
    >>> kspace = torch.randn(1, 1, 640, 368)
    >>> mask, acceleration = mask_func(kspace.shape)
    >>> mask.shape
    torch.Size([1, 1, 640, 368])
    >>> acceleration
    4
    """

    def __call__(
        self,
        shape: Union[Sequence[int], np.ndarray],
        seed: Optional[Union[int, Tuple[int, ...]]] = None,
        partial_fourier_percentage: Optional[float] = 0.0,
        center_scale: Optional[float] = 0.02,
        **kwargs,
    ) -> Tuple[torch.Tensor, int]:
        """Calls :class:`Gaussian2DMaskFunc`.

        Parameters
        ----------
        shape : Sequence[int]
            The shape of the mask to be created. The shape should have at least 3 dimensions. Same as the shape of the
            input k-space data.
        seed : int or tuple of ints, optional
            Seed for the random number generator. Default is ``None``.
        partial_fourier_percentage : float, optional
            Percentage of the low-frequency columns to be retained. Default is ``0.0``.
        center_scale : float, optional
            For autocalibration purposes, data points near the k-space center will be fully sampled within an ellipse
            of which the half-axes will be set to the given `center_scale` percentage of the fully sampled region.
            Default is ``0.02``.

        Returns
        -------
        Tuple[torch.Tensor, int]
            A tuple of the generated mask and the acceleration factor.

        Raises
        ------
        ValueError
            If the `shape` parameter has less than 3 dimensions.
        """
        with temp_seed(self.rng, seed):
            dims = [1 for _ in shape]
            self.shape = tuple(shape[-3:-1])
            dims[-3:-1] = self.shape

            full_width_half_maximum, acceleration = self.choose_acceleration()

            self.full_width_half_maximum = full_width_half_maximum
            self.acceleration = acceleration
            self.center_scale = center_scale

            mask = self.gaussian_kspace()
            mask[tuple(self.gaussian_coordinates())] = 1.0

            if partial_fourier_percentage != 0:
                mask[: int(np.round(mask.shape[0] * partial_fourier_percentage)), :] = 0.0

            mask = torch.from_numpy(mask.reshape(dims).astype(np.float32))

        return mask, acceleration

    def gaussian_kspace(self) -> np.ndarray:
        """Creates a Gaussian sampled k-space center."""
        a, b = self.center_scale * self.shape[0], self.center_scale * self.shape[1]
        afocal, bfocal = self.shape[0] / 2, self.shape[1] / 2
        xx, yy = np.mgrid[: self.shape[0], : self.shape[1]]
        ellipse = np.power((xx - afocal) / a, 2) + np.power((yy - bfocal) / b, 2)
        return (ellipse < 1).astype(float)

    def gaussian_coordinates(self) -> List[Tuple[int, int]]:
        r"""Returns Gaussian sampled k-space coordinates.

        Returns
        -------
        xsamples : np.ndarray
            A 1D numpy array of x-coordinates.
        ysamples : np.ndarray
            A 1D numpy array of y-coordinates.

        Notes
        -----
        The number of samples taken is determined by `n_sample` which is calculated as \
        `self.shape[0] / self.acceleration`. The selection of the samples is based on the probabilities calculated \
        from `gaussian_kernel`.
        """
        n_sample = int(self.shape[0] * self.shape[1] / self.acceleration)
        cartesian_prod = list(np.ndindex(self.shape))
        kernel = self.gaussian_kernel()
        idxs = np.random.choice(range(len(cartesian_prod)), size=n_sample, replace=False, p=kernel.flatten())
        return list(zip(*list(map(cartesian_prod.__getitem__, idxs))))

    def gaussian_kernel(self) -> np.ndarray:
        r"""Creates a Gaussian sampled k-space kernel.

        .. note::
            The function calculates the Gaussian kernel by computing the sum of the exponential of the squared \
            x-values divided by 2 times the square of the standard deviation. The standard deviation is calculated \
            from the full width at half maximum (FWHM) of the Gaussian curve and is defined as the FWHM divided by \
            the square root of 8 times the natural logarithm of 2. The FWHM and the kern_len are obtained from the \
            `full_width_half_maximum` and `shape` attributes of the class respectively.

        Returns
        -------
        ndarray
            The Gaussian kernel.
        """
        kernels = []
        for kern_len in self.shape:
            sigma = self.full_width_half_maximum / np.sqrt(8 * np.log(2))
            x = np.linspace(-1.0, 1.0, kern_len)
            g = np.exp(-(x**2 / (2 * sigma**2)))
            kernels.append(g)
        kernel = np.sqrt(np.outer(kernels[0], kernels[1]))
        kernel = kernel / kernel.sum()
        return kernel


class Poisson2DMaskFunc(MaskFunc):
    r"""Generate variable-density Poisson-disc sampling pattern, as described in [1]_.

    The function generates a variable density Poisson-disc sampling mask with density proportional to
    :math:`1 / (1 + s |r|)`, where :math:`r` represents the k-space radius, and :math:`s` represents the slope. A
    binary search is performed on the slope :math:`s` such that the resulting acceleration factor is close to the
    prescribed acceleration factor `accel`. The parameter `tol` determines how much they can deviate.

    References
    ----------
    .. [1] Bridson, Robert. "Fast Poisson disk sampling in arbitrary dimensions." SIGGRAPH sketches. 2007

    .. note::
        Taken and adapted from: https://github.com/mikgroup/sigpy/blob/master/sigpy/mri/samp.py
    """

    def __call__(
        self,
        shape: Union[Sequence[int], np.ndarray],
        seed: Optional[Union[int, Tuple[int, ...]]] = None,
        partial_fourier_percentage: Optional[float] = 0.0,
        center_scale: Optional[float] = 0.02,
        calib: Optional[Tuple[float, float]] = (0.0, 0.0),
        crop_corner: bool = True,
        max_attempts: int = 30,
        tol: float = 0.3,
        **kwargs,
    ) -> Tuple[torch.Tensor, int]:
        """Calls :class:`Poisson2DMaskFunc`.

        Parameters
        ----------
        shape : Sequence[int]
            The shape of the mask to be created. The shape should have at least 3 dimensions. Same as the shape of the
            input k-space data.
        seed : int or tuple of ints, optional
            Seed for the random number generator. Default is ``None``.
        partial_fourier_percentage : float, optional
            Percentage of the low-frequency columns to be retained. Default is ``0.0``.
        center_scale : float, optional
            For autocalibration purposes, data points near the k-space center will be fully sampled within an ellipse
            of which the half-axes will be set to the given `center_scale` percentage of the fully sampled region.
            Default is ``0.02``.
        calib : Optional[Tuple[float, float]], optional
            Defines the size of the calibration region, which is a square region in the center of k-space. The first
            value defines the percentage of the center that is sampled, and the second value defines the size of the
            calibration region in the center of k-space. Default is ``(0.0, 0.0)``.
        crop_corner : bool, optional
            If set to True, the center of the mask will be cropped to the size of the calibration region. Default is
            ``True``.
        max_attempts : int, optional
            Maximum number of attempts to generate a mask with the desired acceleration factor. Default is ``30``.
        tol : float, optional
            Tolerance for the acceleration factor. Default is ``0.3``.

        Returns
        -------
        Tuple[torch.Tensor, int]
            A tuple containing the mask and the acceleration factor.
            The mask is a `torch.Tensor` of the same shape as the input shape.
            The acceleration factor is an `int` that represents the number of samples taken in the k-space.
        """
        with temp_seed(self.rng, seed):
            self.shape = tuple(shape[-3:-1])
            self.center_scale = center_scale
            _, self.acceleration = self.choose_acceleration()

            ny, nx = self.shape
            y, x = np.mgrid[:ny, :nx]
            x = np.maximum(abs(x - nx / 2) - calib[-1] / 2, 0)  # type: ignore
            x /= x.max()
            y = np.maximum(abs(y - ny / 2) - calib[-2] / 2, 0)  # type: ignore
            y /= y.max()

            r = np.hypot(x, y)

            slope_max = 40.0
            slope_min = 0.0

            d = max(nx, ny)

            while slope_min < slope_max:
                slope = (slope_max + slope_min) / 2
                radius_x = np.clip((1 + r * slope) * nx / d, 1, None)
                radius_y = np.clip((1 + r * slope) * ny / d, 1, None)

                mask = self.generate_poisson_mask(nx, ny, max_attempts, radius_x, radius_y, calib)

                if crop_corner:
                    mask *= r < 1

                with np.errstate(divide="ignore", invalid="ignore"):
                    actual_acceleration = mask.size / np.sum(mask)

                if abs(actual_acceleration - self.acceleration) < tol:
                    break
                if actual_acceleration < self.acceleration:
                    slope_min = slope
                else:
                    slope_max = slope

            pattern1 = mask
            pattern2 = self.centered_circle()
            mask = np.logical_or(pattern1, pattern2)

            if abs(actual_acceleration - self.acceleration) >= tol:
                raise ValueError(f"Cannot generate mask to satisfy acceleration factor of {self.acceleration}.")

            if partial_fourier_percentage != 0:
                mask[:, : int(np.round(mask.shape[1] * partial_fourier_percentage))] = 0.0

            mask = torch.from_numpy(mask.reshape(self.shape).astype(np.float32)).unsqueeze(0).unsqueeze(-1)

        return mask, self.acceleration

    def centered_circle(self) -> np.ndarray:
        """Creates a boolean centered circle image using the center_scale as a radius.

        Returns
        -------
        np.ndarray
            A 2D array of type bool, where True values indicate the points inside the centered circle and False
            values indicate the points outside the centered circle. The circle has its center at the center of the
            input shape and its radius is determined by the `center_scale` attribute.
        """
        center_x = int((self.shape[0] - 1) / 2)
        center_y = int((self.shape[1] - 1) / 2)

        X, Y = np.indices(self.shape)
        radius = int(self.shape[0] * self.center_scale)
        radius_squared = radius**2
        return ((X - center_x) ** 2 + (Y - center_y) ** 2) < radius_squared

    @staticmethod
    @nb.jit(nopython=True, cache=True)
    def generate_poisson_mask(
        nx: int,
        ny: int,
        max_attempts: int,
        radius_x: np.ndarray,
        radius_y: np.ndarray,
        calib: Tuple[float, float],
    ) -> np.ndarray:
        """Generates a Poisson mask of shape `(ny, nx)` by placing points on the grid according to a Poisson
        distribution.

        Parameters
        ----------
        nx : int
            The number of columns in the output mask.
        ny : int
            The number of rows in the output mask.
        max_attempts : int
            The maximum number of attempts to generate a mask with the desired acceleration factor.
        radius_x : np.ndarray
            An array of shape `(ny, nx)` representing the radius of the Poisson distribution in the x-direction.
        radius_y : np.ndarray
            An array of shape `(ny, nx)` representing the radius of the Poisson distribution in the y-direction.
        calib : Tuple[float, float]
            Defines the size of the calibration region. The calibration region is a square region in the center of
            k-space. The first value defines the percentage of the center that is sampled. The second value defines
            the size of the calibration region in the center of k-space.

        Returns
        -------
        np.ndarray
            A binary mask with points placed according to a Poisson distribution.
        """
        mask = np.zeros((ny, nx))

        # Add calibration region
        mask[
            int(ny / 2 - calib[-2] / 2) : int(ny / 2 + calib[-2] / 2),
            int(nx / 2 - calib[-1] / 2) : int(nx / 2 + calib[-1] / 2),
        ] = 1

        # initialize active list
        pxs = np.empty(nx * ny, np.int32)
        pys = np.empty(nx * ny, np.int32)
        pxs[0] = np.random.randint(0, nx)
        pys[0] = np.random.randint(0, ny)
        num_actives = 1
        while num_actives > 0:
            i = np.random.randint(0, num_actives)
            px = pxs[i]
            py = pys[i]
            rx = radius_x[py, px]
            ry = radius_y[py, px]

            # Attempt to generate point
            done = False
            k = 0
            while not done and k < max_attempts:
                # Generate point randomly from r and 2 * r
                v = (np.random.random() * 3 + 1) ** 0.5
                t = 2 * np.pi * np.random.random()
                qx = px + v * rx * np.cos(t)
                qy = py + v * ry * np.sin(t)

                # Reject if outside grid or close to other points
                if 0 <= qx < nx and 0 <= qy < ny:
                    startx = max(int(qx - rx), 0)
                    endx = min(int(qx + rx + 1), nx)
                    starty = max(int(qy - ry), 0)
                    endy = min(int(qy + ry + 1), ny)

                    done = True
                    for x in range(startx, endx):
                        for y in range(starty, endy):
                            if mask[y, x] == 1 and (
                                ((qx - x) / radius_x[y, x]) ** 2 + ((qy - y) / (radius_y[y, x])) ** 2 < 1
                            ):
                                done = False
                                break

                k += 1

            # Add point if done else remove from active list
            if done:
                pxs[num_actives] = qx
                pys[num_actives] = qy
                mask[int(qy), int(qx)] = 1
                num_actives += 1
            else:
                pxs[i] = pxs[num_actives - 1]
                pys[i] = pys[num_actives - 1]
                num_actives -= 1

        return mask


class Random1DMaskFunc(MaskFunc):
    r"""Random1DMaskFunc creates a sub-sampling mask of a given shape.

    The mask selects a subset of columns from the input k-space data. If the k-space data has N columns, the mask
    picks out:

        1. N_low_freqs = (N * center_fraction) columns in the center corresponding to low-frequencies.

        2. The other columns are selected uniformly at random with a probability equal to:
        prob = (N / acceleration - N_low_freqs) /  (N - N_low_freqs). This ensures that the expected number of columns
        selected is equal to (N / acceleration).

    It is possible to use multiple center_fractions and accelerations, in which case one possible (center_fraction,
    acceleration) is chosen uniformly at random each time the Random1DMaskFunc object is called.

    For example, if accelerations = [4, 8] and center_fractions = [0.08, 0.04], then there is a 50% probability that
    4-fold acceleration with 8% center fraction is selected and a 50% probability that 8-fold acceleration with 4%
    center fraction is selected.
    """

    def __call__(
        self,
        shape: Sequence[int],
        seed: Optional[Union[int, Tuple[int, ...]]] = None,
        partial_fourier_percentage: Optional[float] = 0.0,
        **kwargs,
    ) -> Tuple[torch.Tensor, int]:
        """Calls :class:`Random1DMaskFunc`.

        Parameters
        ----------
        shape : Sequence[int]
            The shape of the mask to be created. The shape should have at least 3 dimensions. Same as the shape of the
            input k-space data.
        seed : int or tuple of ints, optional
            Seed for the random number generator. Default is ``None``.
        partial_fourier_percentage : float, optional
            Percentage of the low-frequency columns to be retained. Default is ``0.0``.

        Returns
        -------
        Tuple[torch.Tensor, int]
            A tuple of the generated mask and the acceleration factor.

        Raises
        ------
        ValueError
            If the `shape` parameter has less than 3 dimensions.
        """
        if len(shape) < 3:
            raise ValueError("Shape should have 3 or more dimensions")

        with temp_seed(self.rng, seed):
            num_cols = shape[-2]
            center_fraction, acceleration = self.choose_acceleration()

            # create the mask
            num_low_freqs = int(round(num_cols * center_fraction))
            prob = (num_cols / acceleration - num_low_freqs) / (num_cols - num_low_freqs)
            mask = self.rng.uniform(size=num_cols) < prob
            pad = torch.div((num_cols - num_low_freqs + 1), 2, rounding_mode="trunc").item()
            mask[pad : pad + num_low_freqs] = True

            # reshape the mask
            mask_shape = [1 for _ in shape]
            mask_shape[-2] = num_cols
            mask = torch.from_numpy(mask.reshape(*mask_shape).astype(np.float32))

            if partial_fourier_percentage != 0:
                mask[:, : int(np.round(mask.shape[1] * partial_fourier_percentage))] = 0.0

        return mask, acceleration


def create_masker(
    mask_type_str: str, center_fractions: Union[Sequence[float], float], accelerations: Union[Sequence[int], int]
) -> MaskFunc:
    """Creates a MaskFunc object based on the specified mask type.

    Parameters
    ----------
    mask_type_str : str
        The string representation of the mask type. Must be one of the following:
        'equispaced1d', 'equispaced2d', 'gaussian1d', 'gaussian2d', 'poisson2d', 'random1d'.
    center_fractions : Sequence[float] or float
        The center fractions for the mask.
    accelerations : Sequence[int] or int
        The accelerations for the mask.

    Returns
    -------
    MaskFunc
        A MaskFunc object that corresponds to the specified mask type.

    Raises
    ------
    NotImplementedError
        If the specified `mask_type_str` is not supported.

    Examples
    --------
    >>> from atommic.collections.common.data.subsample import create_masker
    >>> create_masker("random1d", [0.5], [4])
    Random1DMaskFunc([0.5], [4])
    >>> create_masker("equispaced2d", [0.3, 0.7], [8, 6])
    Equispaced2DMaskFunc([0.3, 0.7], [8, 6])
    >>> create_masker("poisson2d", [0.3, 0.7], [8, 6])
    Poisson2DMaskFunc([0.3, 0.7], [8, 6])
    >>> create_masker("gaussian1d", [0.3, 0.7], [8, 6])
    Gaussian1DMaskFunc([0.3, 0.7], [8, 6])
    >>> create_masker("gaussian2d", [0.3, 0.7], [8, 6])
    Gaussian2DMaskFunc([0.3, 0.7], [8, 6])
    """
    if isinstance(center_fractions, float):
        center_fractions = [center_fractions]
    if isinstance(accelerations, int):
        accelerations = [accelerations]
    if mask_type_str == "random1d":
        return Random1DMaskFunc(center_fractions, accelerations)
    if mask_type_str == "equispaced1d":
        return Equispaced1DMaskFunc(center_fractions, accelerations)
    if mask_type_str == "equispaced2d":
        return Equispaced2DMaskFunc(center_fractions, accelerations)
    if mask_type_str == "gaussian1d":
        return Gaussian1DMaskFunc(center_fractions, accelerations)
    if mask_type_str == "gaussian2d":
        return Gaussian2DMaskFunc(center_fractions, accelerations)
    if mask_type_str == "poisson2d":
        return Poisson2DMaskFunc(center_fractions, accelerations)
    raise NotImplementedError(f"{mask_type_str} not supported")



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.quantitative.nn.qcirim import qCIRIM  # noqa: F401
from atommic.collections.quantitative.nn.qvarnet import qVarNet  # noqa: F401



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.utils.callbacks.atommic_model_checkpoint import ATOMMICModelCheckpoint  # noqa: F401
from atommic.utils.callbacks.preemption import PreemptionCallback  # noqa: F401



# coding=utf-8

from atommic.collections.multitask import rs  # noqa: F401
from atommic.package_info import __version__

# Set collection version equal to atommic version.
__version = __version__

# Authorship.
__author__ = "Dimitris Karkalousos"

# Set collection name.
__description__ = "MRI multitask learning collection."



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/optim/radam.py

import math

import torch
from torch.optim.optimizer import Optimizer


class RAdam(Optimizer):
    """RAdam optimizer."""

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=0):
        """Inits :class:`RAdam`.

        Parameters
        ----------
        params : iterable
            Iterable of parameters to optimize or dicts defining parameter groups.
        lr : float, optional
            Learning rate. Default is ``1e-3``.
        betas : tuple of floats, optional
            Coefficients used for computing running averages of gradient and its square. Default is ``(0.9, 0.999)``.
        eps : float, optional
            Term added to the denominator to improve numerical stability. Default is ``1e-8``.
        weight_decay : float, optional
            Weight decay (L2 penalty). Default is ``0``.
        """
        defaults = {"lr": lr, "betas": betas, "eps": eps, "weight_decay": weight_decay}
        self.buffer = [[None, None, None] for _ in range(10)]
        super().__init__(params, defaults)

    def step(self, closure=None):
        """Step through the optimizer"""
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data.float()
                if grad.is_sparse:
                    raise RuntimeError("RAdam does not support sparse gradients")

                p_data_fp32 = p.data.float()

                state = self.state[p]

                if len(state) == 0:
                    state["step"] = 0
                    state["exp_avg"] = torch.zeros_like(p_data_fp32)
                    state["exp_avg_sq"] = torch.zeros_like(p_data_fp32)
                else:
                    state["exp_avg"] = state["exp_avg"].type_as(p_data_fp32)
                    state["exp_avg_sq"] = state["exp_avg_sq"].type_as(p_data_fp32)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=(1.0 - beta2))
                exp_avg.mul_(beta1).add_(grad, alpha=(1.0 - beta1))

                state["step"] += 1
                buffered = self.buffer[int(state["step"] % 10)]
                if state["step"] == buffered[0]:
                    N_sma, step_size = buffered[1], buffered[2]
                else:
                    buffered[0] = state["step"]
                    beta2_t = beta2 ** state["step"]
                    N_sma_max = 2 / (1 - beta2) - 1
                    N_sma = N_sma_max - 2 * state["step"] * beta2_t / (1 - beta2_t)
                    buffered[1] = N_sma

                    # more conservative since it's an approximated value
                    if N_sma >= 5:
                        step_size = (
                            group["lr"]
                            * math.sqrt(
                                (1 - beta2_t)
                                * (N_sma - 4)
                                / (N_sma_max - 4)
                                * (N_sma - 2)
                                / N_sma
                                * N_sma_max
                                / (N_sma_max - 2)
                            )
                            / (1 - beta1 ** state["step"])
                        )
                    else:
                        step_size = group["lr"] / (1 - beta1 ** state["step"])
                    buffered[2] = step_size

                if group["weight_decay"] != 0:
                    p_data_fp32.add_(-group["weight_decay"] * group["lr"], p_data_fp32)

                # more conservative since it's an approximated value
                if N_sma >= 5:
                    denom = exp_avg_sq.sqrt().add_(group["eps"])
                    p_data_fp32.addcdiv_(-step_size, exp_avg, denom)
                else:
                    p_data_fp32.add_(-step_size, exp_avg)

                p.data.copy_(p_data_fp32)

        return loss



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NKI-AI/direct/blob/main/direct/nn/didn/didn.py

import torch
import torch.nn.functional as F
from torch import nn


class Subpixel(nn.Module):
    """Subpixel convolution layer for up-scaling of low resolution features at super-resolution as presented in
    [Songhyun2019]_.

    References
    ----------
    .. [Songhyun2019] Yu, Songhyun, et al. “Deep Iterative Down-Up CNN for Image Denoising.” 2019 IEEE/CVF Conference
        on Computer Vision and Pattern Recognition Workshops (CVPRW), 2019, pp. 2095–103. IEEE Xplore,
        https://doi.org/10.1109/CVPRW.2019.00262.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        upscale_factor: int,
        kernel_size: int,
        padding: int = 0,
    ):
        """Inits :class:`Subpixel`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        upscale_factor : int
            Subpixel upscale factor.
        kernel_size : int
            Convolution kernel size.
        padding : int, optional
            Padding size. Default is ``0``.
        """
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels, out_channels * upscale_factor**2, kernel_size=kernel_size, padding=padding
        )
        self.pixelshuffle = nn.PixelShuffle(upscale_factor)

    def forward(self, x):
        """Computes Subpixel convolution."""
        return self.pixelshuffle(self.conv(x))


class ReconBlock(nn.Module):
    """Reconstruction Block of DIDN model as presented in [Songhyun2019]_.

    References
    ----------
    .. [Songhyun2019] Yu, Songhyun, et al. “Deep Iterative Down-Up CNN for Image Denoising.” 2019 IEEE/CVF Conference
        on Computer Vision and Pattern Recognition Workshops (CVPRW), 2019, pp. 2095–103. IEEE Xplore,
        https://doi.org/10.1109/CVPRW.2019.00262.
    """

    def __init__(self, in_channels: int, num_convs: int):
        """Inits :class:`ReconBlock`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        num_convs : int
            Number of convolution blocks.
        """
        super().__init__()
        self.convs = nn.ModuleList(
            [
                nn.Sequential(
                    *[
                        nn.Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=3, padding=1),
                        nn.PReLU(),
                    ]
                )
                for _ in range(num_convs - 1)
            ]
        )
        self.convs.append(nn.Conv2d(in_channels=in_channels, out_channels=in_channels, kernel_size=3, padding=1))
        self.num_convs = num_convs

    def forward(self, input_data: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`ReconBlock`.

        Parameters
        ----------
        input_data : torch.Tensor
            Input data.
        """
        output = input_data.clone()
        for idx in range(self.num_convs):
            output = self.convs[idx](output)

        return input_data + output


class DUB(nn.Module):
    r"""Down-Up Block of DIDN model as presented in [Songhyun2019]_.

    References
    ----------
    .. [Songhyun2019] Yu, Songhyun, et al. “Deep Iterative Down-Up CNN for Image Denoising.” 2019 IEEE/CVF Conference
        on Computer Vision and Pattern Recognition Workshops (CVPRW), 2019, pp. 2095–103. IEEE Xplore,
        https://doi.org/10.1109/CVPRW.2019.00262.

    """

    def __init__(self, in_channels: int, out_channels: int):
        """Inits :class:`DUB`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        """
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels

        # Scale 1
        self.conv1_1 = nn.Sequential(*[nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=1), nn.PReLU()] * 2)
        self.down1 = nn.Conv2d(in_channels, in_channels * 2, kernel_size=3, stride=2, padding=1)
        # Scale 2
        self.conv2_1 = nn.Sequential(
            *[nn.Conv2d(in_channels * 2, in_channels * 2, kernel_size=3, padding=1), nn.PReLU()]
        )
        self.down2 = nn.Conv2d(in_channels * 2, in_channels * 4, kernel_size=3, stride=2, padding=1)
        # Scale 3
        self.conv3_1 = nn.Sequential(
            *[
                nn.Conv2d(in_channels * 4, in_channels * 4, kernel_size=3, padding=1),
                nn.PReLU(),
            ]
        )
        self.up1 = nn.Sequential(*[Subpixel(in_channels * 4, in_channels * 2, 2, 1, 0)])
        # Scale 2
        self.conv_agg_1 = nn.Conv2d(in_channels * 4, in_channels * 2, kernel_size=1)
        self.conv2_2 = nn.Sequential(
            *[
                nn.Conv2d(in_channels * 2, in_channels * 2, kernel_size=3, padding=1),
                nn.PReLU(),
            ]
        )
        self.up2 = nn.Sequential(*[Subpixel(in_channels * 2, in_channels, 2, 1, 0)])
        # Scale 1
        self.conv_agg_2 = nn.Conv2d(in_channels * 2, in_channels, kernel_size=1)
        self.conv1_2 = nn.Sequential(*[nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=1), nn.PReLU()] * 2)
        self.conv_out = nn.Sequential(*[nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=1), nn.PReLU()])

    @staticmethod
    def pad(x: torch.Tensor) -> torch.Tensor:
        """Pads input to height and width dimensions if odd.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            Padded tensor.
        """
        padding = [0, 0, 0, 0]

        if x.shape[-2] % 2 != 0:
            padding[3] = 1  # Padding right - width
        if x.shape[-1] % 2 != 0:
            padding[1] = 1  # Padding bottom - height
        if sum(padding) != 0:
            x = F.pad(x, padding, "reflect")
        return x

    @staticmethod
    def crop_to_shape(x: torch.Tensor, shape: tuple) -> torch.Tensor:
        r"""Crops ``x`` to specified shape.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor with shape ('\'*, H, W).
        shape : tuple
            Crop shape corresponding to H, W.

        Returns
        -------
        torch.Tensor
            Cropped tensor.
        """
        h, w = x.shape[-2:]
        if h > shape[0]:
            x = x[:, :, : shape[0], :]
        if w > shape[1]:
            x = x[:, :, :, : shape[1]]
        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`DUB`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            DUB output.
        """
        x1 = self.pad(x.clone())
        x1 = x1 + self.conv1_1(x1)
        x2 = self.down1(x1)
        x2 = x2 + self.conv2_1(x2)
        out = self.down2(x2)
        out = out + self.conv3_1(out)
        out = self.up1(out)
        out = torch.cat([x2, self.crop_to_shape(out, x2.shape[-2:])], dim=1)
        out = self.conv_agg_1(out)
        out = out + self.conv2_2(out)
        out = self.up2(out)
        out = torch.cat([x1, self.crop_to_shape(out, x1.shape[-2:])], dim=1)
        out = self.conv_agg_2(out)
        out = out + self.conv1_2(out)
        out = x + self.crop_to_shape(self.conv_out(out), x.shape[-2:])
        return out


class DIDN(nn.Module):
    r"""Deep Iterative Down-up convolutional Neural network (DIDN), as presented in [Songhyun2019]_.

    References
    ----------
    .. [Songhyun2019] Yu, Songhyun, et al. “Deep Iterative Down-Up CNN for Image Denoising.” 2019 IEEE/CVF Conference
        on Computer Vision and Pattern Recognition Workshops (CVPRW), 2019, pp. 2095–103. IEEE Xplore,
        https://doi.org/10.1109/CVPRW.2019.00262.

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        hidden_channels: int = 128,
        num_dubs: int = 6,
        num_convs_recon: int = 9,
        skip_connection: bool = False,
    ):
        """Inits :class:`DIDN`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        hidden_channels : int, optional
            Number of hidden channels. First convolution out_channels. Default: 128.
        num_dubs : int, optional
            Number of DUB networks. Default: 6.
        num_convs_recon : int, optional
            Number of ReconBlock convolutions. Default: 9.
        skip_connection : bool, optional
            Use skip connection. Default: False.
        """
        super().__init__()
        self.conv_in = nn.Sequential(
            *[nn.Conv2d(in_channels=in_channels, out_channels=hidden_channels, kernel_size=3, padding=1), nn.PReLU()]
        )
        self.down = nn.Conv2d(
            in_channels=hidden_channels,
            out_channels=hidden_channels,
            kernel_size=3,
            stride=2,
            padding=1,
        )
        self.dubs = nn.ModuleList(
            [DUB(in_channels=hidden_channels, out_channels=hidden_channels) for _ in range(num_dubs)]
        )
        self.recon_block = ReconBlock(in_channels=hidden_channels, num_convs=num_convs_recon)
        self.recon_agg = nn.Conv2d(in_channels=hidden_channels * num_dubs, out_channels=hidden_channels, kernel_size=1)
        self.conv = nn.Sequential(
            *[
                nn.Conv2d(
                    in_channels=hidden_channels,
                    out_channels=hidden_channels,
                    kernel_size=3,
                    padding=1,
                ),
                nn.PReLU(),
            ]
        )
        self.up2 = Subpixel(hidden_channels, hidden_channels, 2, 1)
        self.conv_out = nn.Conv2d(
            in_channels=hidden_channels,
            out_channels=out_channels,
            kernel_size=3,
            padding=1,
        )
        self.num_dubs = num_dubs
        self.skip_connection = (in_channels == out_channels) and skip_connection

    @staticmethod
    def crop_to_shape(x: torch.Tensor, shape: tuple) -> torch.Tensor:
        r"""Crops ``x`` to specified shape.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor with shape ('\'*, H, W).
        shape : tuple
            Crop shape corresponding to H, W.

        Returns
        -------
        torch.Tensor
            Cropped tensor.
        """
        h, w = x.shape[-2:]

        if h > shape[0]:
            x = x[:, :, : shape[0], :]
        if w > shape[1]:
            x = x[:, :, :, : shape[1]]
        return x

    def forward(self, x: torch.Tensor, channel_dim: int = 1) -> torch.Tensor:
        r"""Forward pass of :class:`DIDN`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor with shape ('\'*, C, H, W).
        channel_dim : int, optional
            Channel dimension. Default is ``1``.

        Returns
        -------
        torch.Tensor
            Output tensor with shape ('\'*, C, H, W).
        """
        out = self.conv_in(x)
        out = self.down(out)

        dub_outs = []
        for dub in self.dubs:
            out = dub(out)
            dub_outs.append(out)

        out = [self.recon_block(dub_out) for dub_out in dub_outs]
        out = self.recon_agg(torch.cat(out, dim=channel_dim))
        out = self.conv(out)
        out = self.up2(out)
        out = self.conv_out(out)
        out = self.crop_to_shape(out, x.shape[-2:])

        if self.skip_connection:
            out = x + out
        return out



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import List, Sequence

import torch

from atommic.collections.common.parts import fft, utils
from atommic.collections.quantitative.nn.base import SignalForwardModel


def expand_op(x: torch.Tensor, sensitivity_maps: torch.Tensor) -> torch.Tensor:
    """Expands a coil-combined image to multicoil.

    Parameters
    ----------
    x : torch.Tensor
        Coil-combined image.
    sensitivity_maps : torch.Tensor
        Coil sensitivity maps.

    Returns
    -------
    torch.Tensor
        Multicoil image.

    Examples
    --------
    >>> import torch
    >>> from atommic.collections.quantitative.nn.qrim_base.utils import expand_op
    >>> data = torch.randn(1, 1, 320, 320, 2)
    >>> coil_sensitivity_maps = torch.randn(1, 32, 320, 320, 2)
    >>> expand_op(data, coil_sensitivity_maps).shape
    torch.Size([1, 32, 320, 320, 2])
    """
    x = utils.complex_mul(x, sensitivity_maps)
    if torch.isnan(x).any():
        x = torch.where(torch.isnan(x), torch.zeros_like(x), x)
    return x


def analytical_log_likelihood_gradient(
    linear_forward_model: SignalForwardModel,
    prediction: torch.Tensor,
    TEs: List,
    sensitivity_maps: torch.Tensor,
    masked_kspace: torch.Tensor,
    sampling_mask: torch.Tensor,
    fft_centered: bool,
    fft_normalization: str,
    spatial_dims: Sequence[int],
    coil_dim: int,
    coil_combination_method: str = "SENSE",
    scaling: float = 1e-3,
) -> torch.Tensor:
    """Computes the analytical gradient of the log-likelihood function.

    Parameters
    ----------
    linear_forward_model: SignalForwardModel
        Signal forward model to use.
    prediction : torch.Tensor
        Current prediction of the quantitative maps.
    TEs : List
        List of echo times.
    sensitivity_maps : torch.Tensor
        Coil sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2].
    masked_kspace : torch.Tensor
        Data of shape [batch_size, n_echoes, n_coils, n_x, n_y, 2].
    sampling_mask : torch.Tensor
        Mask of the sampling of shape [batch_size, 1, n_x, n_y, 1].
    fft_centered : bool
        If True, the FFT is centered.
    fft_normalization : str
        Normalization of the FFT.
    spatial_dims : Sequence[int]
        Spatial dimensions of the input.
    coil_dim : int
        Coil dimension.
    coil_combination_method : str, optional
        Coil combination method. Default is ``"SENSE"``.
    scaling : float, optional
        Scaling factor. Default is ``1e-3``.

    Returns
    -------
    torch.Tensor
        Analytical gradient of the log-likelihood function.
    """
    nr_TEs = len(TEs)

    R2star_map, S0_map, B0_map, phi_map = prediction[0], prediction[1], prediction[2], prediction[3]

    R2star_map = R2star_map.unsqueeze(0)
    S0_map = S0_map.unsqueeze(0)
    B0_map = B0_map.unsqueeze(0)
    phi_map = phi_map.unsqueeze(0)

    pred = linear_forward_model(R2star_map, S0_map, B0_map, phi_map, TEs)

    S0_map_real = S0_map
    S0_map_imag = phi_map

    pred_kspace = fft.fft2(
        expand_op(pred.unsqueeze(coil_dim), sensitivity_maps.unsqueeze(0).unsqueeze(coil_dim - 1)),
        fft_centered,
        fft_normalization,
        spatial_dims,
    )

    diff_data = (pred_kspace - masked_kspace) * sampling_mask
    diff_data_inverse = utils.coil_combination_method(
        fft.ifft2(diff_data, fft_centered, fft_normalization, spatial_dims),
        sensitivity_maps.unsqueeze(0).unsqueeze(coil_dim - 1),
        method=coil_combination_method,
        dim=coil_dim,
    )

    def first_term(i):
        """First term of the gradient."""
        return torch.exp(-TEs[i] * scaling * R2star_map)

    def second_term(i):
        """Second term of the gradient."""
        return torch.cos(B0_map * scaling * -TEs[i])

    def third_term(i):
        """Third term of the gradient."""
        return torch.sin(B0_map * scaling * -TEs[i])

    S0_part_der = torch.stack(
        [torch.stack((first_term(i) * second_term(i), -first_term(i) * third_term(i)), -1) for i in range(nr_TEs)], 1
    )

    R2str_part_der = torch.stack(
        [
            torch.stack(
                (
                    -TEs[i] * scaling * first_term(i) * (S0_map_real * second_term(i) - S0_map_imag * third_term(i)),
                    -TEs[i] * scaling * first_term(i) * (-S0_map_real * third_term(i) - S0_map_imag * second_term(i)),
                ),
                -1,
            )
            for i in range(nr_TEs)
        ],
        1,
    )

    S0_map_real_grad = (
        diff_data_inverse[..., 0] * S0_part_der[..., 0] - diff_data_inverse[..., 1] * S0_part_der[..., 1]
    )
    S0_map_imag_grad = (
        diff_data_inverse[..., 0] * S0_part_der[..., 1] + diff_data_inverse[..., 1] * S0_part_der[..., 0]
    )
    R2star_map_real_grad = (
        diff_data_inverse[..., 0] * R2str_part_der[..., 0] - diff_data_inverse[..., 1] * R2str_part_der[..., 1]
    )
    R2star_map_imag_grad = (
        diff_data_inverse[..., 0] * R2str_part_der[..., 1] + diff_data_inverse[..., 1] * R2str_part_der[..., 0]
    )

    S0_map_grad = torch.stack([S0_map_real_grad, S0_map_imag_grad], -1).squeeze()
    S0_map_grad = torch.mean(S0_map_grad, 0)
    R2star_map_grad = torch.stack([R2star_map_real_grad, R2star_map_imag_grad], -1).squeeze()
    R2star_map_grad = torch.mean(R2star_map_grad, 0)

    return torch.stack([R2star_map_grad[..., 0], S0_map_grad[..., 0], R2star_map_grad[..., 1], S0_map_grad[..., 1]], 0)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from atommic.collections.segmentation.data.mri_segmentation_loader import (  # noqa: F401
    BraTS2023AdultGliomaSegmentationMRIDataset,
    ISLES2022SubAcuteStrokeSegmentationMRIDataset,
    SegmentationMRIDataset,
    SKMTEASegmentationMRIDataset,
)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Dict, List, Union

import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import check_stacked_complex, coil_combination_method, expand_op
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.crnn_base.crnn_block import GRUConv2d, RecurrentConvolutionalNetBlock
from atommic.core.classes.common import typecheck

__all__ = ["CRNNet"]


class CRNNet(BaseMRIReconstructionModel):
    """Implementation of the Convolutional Recurrent Neural Network, as presented in [Qin2019]_.

    References
    ----------
    .. [Qin2019] C. Qin, J. Schlemper, J. Caballero, A. N. Price, J. V. Hajnal and D. Rueckert, "Convolutional
        Recurrent Neural Networks for Dynamic MR Image Reconstruction," in IEEE Transactions on Medical Imaging, vol.
        38, no. 1, pp. 280-290, Jan. 2019, doi: 10.1109/TMI.2018.2863670.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`CRNNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.no_dc = cfg_dict.get("no_dc")
        self.num_iterations = cfg_dict.get("num_iterations")

        self.reconstruction_module = RecurrentConvolutionalNetBlock(
            GRUConv2d(
                in_channels=2,
                out_channels=2,
                hidden_channels=cfg_dict.get("hidden_channels"),
                n_convs=cfg_dict.get("n_convs"),
                batchnorm=cfg_dict.get("batchnorm"),
            ),
            num_iterations=self.num_iterations,
            fft_centered=self.fft_centered,
            fft_normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
            coil_dim=self.coil_dim,
            no_dc=self.no_dc,
        )

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> List[torch.Tensor]:
        """Forward pass of :class:`CRNNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        List of torch.Tensor
            List of the intermediate predictions for each cascade. Shape [batch_size, n_x, n_y].
        """
        predictions = self.reconstruction_module(y, sensitivity_maps, mask)
        return [self.process_intermediate_pred(x, sensitivity_maps) for x in predictions]

    def process_intermediate_pred(
        self, prediction: Union[List, torch.Tensor], sensitivity_maps: torch.Tensor
    ) -> torch.Tensor:
        """Process the intermediate prediction.

        Parameters
        ----------
        prediction : torch.Tensor
            Intermediate prediction. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]

        Returns
        -------
        torch.Tensor
            Processed prediction. Shape [batch_size, n_x, n_y].
        """
        return check_stacked_complex(
            coil_combination_method(
                ifft2(
                    prediction,
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                ),
                sensitivity_maps,
                self.coil_combination_method,
                self.coil_dim,
            )
        )

    def process_reconstruction_loss(
        self,
        target: torch.Tensor,
        prediction: Union[list, torch.Tensor],
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        attrs: Dict,
        r: int,
        loss_func: torch.nn.Module,
    ) -> torch.Tensor:
        """Processes the reconstruction loss for the CRNNet model. It differs from the base class in that it uses the
        intermediate predictions to compute the loss.

        Parameters
        ----------
        target : torch.Tensor
            Target data of shape [batch_size, n_x, n_y, 2].
        prediction : Union[list, torch.Tensor]
            Prediction(s) of shape [batch_size, n_x, n_y, 2].
        sensitivity_maps : torch.Tensor
            Sensitivity maps of shape [batch_size, n_coils, n_x, n_y, 2]. It will be used if self.ssdu is True, to
            expand the target and prediction to multiple coils.
        mask : torch.Tensor
            Mask of shape [batch_size, n_x, n_y, 2]. It will be used if self.ssdu is True, to enforce data consistency
            on the prediction.
        attrs : Dict
            Attributes of the data with pre normalization values.
        r : int
            The selected acceleration factor.
        loss_func : torch.nn.Module
            Loss function. Must be one of {torch.nn.L1Loss(), torch.nn.MSELoss(),
            atommic.collections.reconstruction.losses.ssim.SSIMLoss()}. Default is ``torch.nn.L1Loss()``.

        Returns
        -------
        loss: torch.FloatTensor
            If self.accumulate_loss is True, returns an accumulative result of all intermediate losses.
            Otherwise, returns the loss of the last intermediate loss.
        """
        # If kspace reconstruction loss is used, the target needs to be transformed to k-space.
        if self.kspace_reconstruction_loss:
            # If inputs are complex, then they need to be viewed as real.
            if target.shape[-1] != 2 and torch.is_complex(target):
                target = torch.view_as_real(target)
            # If SSDU is used, then the coil-combined inputs need to be expanded to multiple coils using the
            # sensitivity maps.
            if self.ssdu:
                target = expand_op(target, sensitivity_maps, self.coil_dim)
            # Transform to k-space.
            target = fft2(target, self.fft_centered, self.fft_normalization, self.spatial_dims)
            # Ensure loss inputs are both viewed in the same way.
            target = self.__abs_output__(target)
        elif not self.unnormalize_loss_inputs:
            # Ensure loss inputs are both viewed in the same way.
            target = self.__abs_output__(target)
            # Normalize inputs to [0, 1]
            target = torch.abs(target / torch.max(torch.abs(target)))

        def compute_reconstruction_loss(t, p, s):
            if self.unnormalize_loss_inputs:
                # we do the unnormalization here to avoid explicitly iterating through list of predictions, which
                # might be a list of lists.
                t, p, s = self.__unnormalize_for_loss_or_log__(t, p, s, attrs, r)

            # If kspace reconstruction loss is used, the target needs to be transformed to k-space.
            if self.kspace_reconstruction_loss:
                # If inputs are complex, then they need to be viewed as real.
                if p.shape[-1] != 2 and torch.is_complex(p):
                    p = torch.view_as_real(p)
                # If SSDU is used, then the coil-combined inputs need to be expanded to multiple coils using the
                # sensitivity maps.
                if self.ssdu:
                    p = expand_op(p, s, self.coil_dim)
                # Transform to k-space.
                p = fft2(p, self.fft_centered, self.fft_normalization, self.spatial_dims)
                # If SSDU is used, then apply the mask to the prediction to enforce data consistency.
                if self.ssdu:
                    p = p * mask
                # Ensure loss inputs are both viewed in the same way.
                p = self.__abs_output__(p)
            elif not self.unnormalize_loss_inputs:
                p = self.__abs_output__(p)
                # Normalize inputs to [0, 1]
                p = torch.abs(p / torch.max(torch.abs(p)))

            if "ssim" in str(loss_func).lower():
                return loss_func(
                    t.unsqueeze(dim=self.coil_dim),
                    p.unsqueeze(dim=self.coil_dim),
                    data_range=torch.tensor(
                        [max(torch.max(t).item(), torch.max(p).item()) - min(torch.min(t).item(), torch.min(p).item())]
                    )
                    .unsqueeze(dim=0)
                    .to(t.device),
                )
            return loss_func(t, p)

        iterations_weights = torch.logspace(-1, 0, steps=self.num_iterations).to(target.device)
        if self.num_echoes > 0:
            iterations_loss = [
                torch.mean(
                    torch.stack(
                        [
                            compute_reconstruction_loss(
                                target[echo].unsqueeze(0), iteration_pred[echo].unsqueeze(0), sensitivity_maps
                            )
                            for echo in range(target.shape[0])
                        ]
                    )
                ).to(target)
                for iteration_pred in prediction
            ]
        else:
            iterations_loss = [
                compute_reconstruction_loss(target, iteration_pred, sensitivity_maps) for iteration_pred in prediction
            ]
        loss = sum(x * w for x, w in zip(iterations_loss, iterations_weights)) / self.num_iterations
        return loss



# coding=utf-8
__author__ = "Dimitris Karkalousos"

import torch
from torch import nn


class ConvGRUCellBase(nn.Module):
    """Base class for Convolutional Gated Recurrent Unit (GRU) cells."""

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        conv_dim: int,
        kernel_size: int,
        dilation: int,
        bias: bool = True,
    ):
        """Inits :class:`ConvGRUCellBase`.

        Parameters
        ----------
        input_size : int
            Number of input channels.
        hidden_size : int
            Number of hidden channels.
        conv_dim : int
            Number of dimensions of the convolutional layer.
        kernel_size : int
            Size of the convolutional kernel.
        dilation : int
            Dilation of the convolutional kernel.
        bias : bool
            Whether to use bias. Default is ``True``.
        """
        super().__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        self.conv_dim = conv_dim
        self.conv_class = self.determine_conv_class(conv_dim)

        self.ih = nn.Conv2d(
            input_size,
            3 * hidden_size,
            kernel_size,
            padding=torch.div(dilation * (kernel_size - 1), 2, rounding_mode="trunc").item(),
            dilation=dilation,
            bias=bias,
        )
        self.hh = nn.Conv2d(
            hidden_size,
            3 * hidden_size,
            kernel_size,
            padding=torch.div(dilation * (kernel_size - 1), 2, rounding_mode="trunc").item(),
            dilation=dilation,
            bias=False,
        )

        self.reset_parameters()

    def reset_parameters(self):
        """Initialize parameters following the way proposed in the paper."""
        self.ih.weight.data = self.orthotogonalize_weights(self.ih.weight.data)
        self.hh.weight.data = self.orthotogonalize_weights(self.hh.weight.data)

        if self.bias is True:
            nn.init.zeros_(self.ih.bias)

    @staticmethod
    def orthotogonalize_weights(weights: torch.Tensor, chunks: int = 1) -> torch.Tensor:
        """Orthogonalize the weights of a convolutional layer."""
        return torch.cat([nn.init.orthogonal_(w) for w in weights.chunk(chunks, 0)], 0)

    @staticmethod
    def determine_conv_class(n_dim: int) -> nn.Module:
        """Determine the convolutional class to use."""
        if n_dim == 1:
            return nn.Conv1d
        if n_dim == 2:
            return nn.Conv2d
        if n_dim == 3:
            return nn.Conv3d
        raise NotImplementedError("No convolution of this dimensionality implemented")

    def extra_repr(self):
        """Extra information to be printed when printing the model."""
        s = "{input_size}, {hidden_size}"
        if "bias" in self.__dict__ and self.bias is not True:
            s += ", bias={bias}"
        if "nonlinearity" in self.__dict__ and self.nonlinearity != "tanh":
            s += ", nonlinearity={nonlinearity}"
        return s.format(**self.__dict__)

    def check_forward_input(self, _input: torch.Tensor):
        """Check forward input."""
        if _input.size(1) != self.input_size:
            raise RuntimeError(f"input has inconsistent input_size: got {_input.size(1)}, expected {self.input_size}")

    def check_forward_hidden(self, _input: torch.Tensor, hx: torch.Tensor, hidden_label: str = ""):
        """Check forward hidden."""
        if _input.size(0) != hx.size(0):
            raise RuntimeError(
                f"Input batch size {_input.size(0)} doesn't match hidden{hidden_label} batch size {hx.size(0)}"
            )

        if hx.size(1) != self.hidden_size:
            raise RuntimeError(
                f"hidden{hidden_label} has inconsistent hidden_size: got {hx.size(1)}, expected {self.hidden_size}"
            )


class ConvGRUCell(ConvGRUCellBase):
    """A Convolutional GRU cell."""

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        conv_dim: int,
        kernel_size: int,
        dilation: int,
        bias: bool = True,
    ):
        """Inits :class:`ConvGRUCell`.

        Parameters
        ----------
        input_size : int
            Number of input channels.
        hidden_size : int
            Number of hidden channels.
        conv_dim : int
            Number of dimensions of the convolutional layer.
        kernel_size : int
            Size of the convolutional kernel.
        dilation : int
            Dilation of the convolutional kernel.
        bias : bool
            Whether to use bias. Default is ``True``.
        """
        super().__init__(input_size, hidden_size, conv_dim, kernel_size, dilation, bias)
        self.conv_dim = conv_dim

    def forward(self, _input: torch.Tensor, hx: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`ConvGRUCell`."""
        if self.conv_dim == 3:
            _input = _input.unsqueeze(0)
            hx = hx.permute(1, 0, 2, 3).unsqueeze(0)

        ih = self.ih(_input).chunk(3, 1)
        hh = self.hh(hx).chunk(3, 1)

        r = torch.sigmoid(ih[0] + hh[0])
        z = torch.sigmoid(ih[1] + hh[1])
        n = torch.tanh(ih[2] + r * hh[2])

        hx = n * (1 - z) + z * hx

        return hx


class ConvMGUCellBase(nn.Module):
    """Base class for a Convolutional Minimal Gated Unit cell, as presented in [Zhou2016]_.

    References
    ----------
    .. [Zhou2016] Zhou GB, Wu J, Zhang CL, Zhou ZH. Minimal gated unit for recurrent neural networks. International
        Journal of Automation and Computing. 2016 Jun;13(3):226-34.
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        conv_dim: int,
        kernel_size: int,
        dilation: int,
        bias: bool = True,
    ):
        """Inits :class:`ConvMGUCellBase`.

        Parameters
        ----------
        input_size : int
            Number of input channels.
        hidden_size : int
            Number of hidden channels.
        conv_dim : int
            Number of dimensions of the convolutional layer.
        kernel_size : int
            Size of the convolutional kernel.
        dilation : int
            Dilation of the convolutional kernel.
        bias : bool
            Whether to use bias. Default is ``True``.
        """
        super().__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.bias = bias
        self.conv_dim = conv_dim
        self.conv_class = self.determine_conv_class(conv_dim)

        self.ih = nn.Conv2d(
            input_size,
            2 * hidden_size,
            kernel_size,
            padding=torch.div(dilation * (kernel_size - 1), 2, rounding_mode="trunc").item(),
            dilation=dilation,
            bias=bias,
        )
        self.hh = nn.Conv2d(
            hidden_size,
            2 * hidden_size,
            kernel_size,
            padding=torch.div(dilation * (kernel_size - 1), 2, rounding_mode="trunc").item(),
            dilation=dilation,
            bias=False,
        )

        self.reset_parameters()

    def reset_parameters(self):
        """Reset the parameters."""
        self.ih.weight.data = self.orthotogonalize_weights(self.ih.weight.data)
        self.hh.weight.data = self.orthotogonalize_weights(self.hh.weight.data)

        nn.init.xavier_uniform_(self.ih.weight, nn.init.calculate_gain("relu"))
        nn.init.xavier_uniform_(self.hh.weight)

        if self.bias is True:
            nn.init.zeros_(self.ih.bias)

    @staticmethod
    def orthotogonalize_weights(weights: torch.Tensor, chunks: int = 1) -> torch.Tensor:
        """Orthogonalize the weights."""
        return torch.cat([nn.init.orthogonal_(w) for w in weights.chunk(chunks, 0)], 0)

    @staticmethod
    def determine_conv_class(n_dim: int) -> nn.Module:
        """Determine the convolutional class."""
        if n_dim == 1:
            return nn.Conv1d
        if n_dim == 2:
            return nn.Conv2d
        if n_dim == 3:
            return nn.Conv3d
        raise ValueError(f"Convolution of: {n_dim} dims is not implemented")

    def extra_repr(self):
        """Extra information about the ConvMGUCellBase."""
        s = "{input_size}, {hidden_size}"
        if "bias" in self.__dict__ and self.bias is not True:
            s += ", bias={bias}"
        if "nonlinearity" in self.__dict__ and self.nonlinearity != "tanh":
            s += ", nonlinearity={nonlinearity}"
        return s.format(**self.__dict__)

    def check_forward_input(self, _input: torch.Tensor):
        """Check the forward input."""
        if _input.size(1) != self.input_size:
            raise RuntimeError(f"input has inconsistent input_size: got {_input.size(1)}, expected {self.input_size}")

    def check_forward_hidden(self, _input: torch.Tensor, hx: torch.Tensor, hidden_label: str = ""):
        """Check the forward hidden."""
        if _input.size(0) != hx.size(0):
            raise RuntimeError(
                f"Input batch size {_input.size(0)} doesn't match hidden{hidden_label} batch size {hx.size(0)}"
            )

        if hx.size(1) != self.hidden_size:
            raise RuntimeError(
                f"hidden{hidden_label} has inconsistent hidden_size: got {hx.size(1)}, expected {self.hidden_size}"
            )


class ConvMGUCell(ConvMGUCellBase):
    """Base class for a Convolutional Minimal Gated Unit cell, as presented in [Zhou2016]_.

    References
    ----------
    .. [Zhou2016] Zhou GB, Wu J, Zhang CL, Zhou ZH. Minimal gated unit for recurrent neural networks. International
        Journal of Automation and Computing. 2016 Jun;13(3):226-34.
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        conv_dim: int,
        kernel_size: int,
        dilation: int,
        bias: bool = True,
    ):
        """Inits :class:`ConvMGUCell`.

        Parameters
        ----------
        input_size : int
            Number of input channels.
        hidden_size : int
            Number of hidden channels.
        conv_dim : int
            Number of dimensions of the convolutional layer.
        kernel_size : int
            Size of the convolutional kernel.
        dilation : int
            Dilation of the convolutional kernel.
        bias : bool
            Whether to use bias. Default is ``True``.
        """
        super().__init__(input_size, hidden_size, conv_dim, kernel_size, dilation, bias)
        self.conv_dim = conv_dim

    def forward(self, _input: torch.Tensor, hx: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`ConvMGUCell`."""
        if self.conv_dim == 3:
            _input = _input.unsqueeze(0)
            hx = hx.permute(1, 0, 2, 3).unsqueeze(0)

        ih = self.ih(_input).chunk(2, dim=1)
        hh = self.hh(hx).chunk(2, dim=1)

        f = torch.sigmoid(ih[0] + hh[0])
        c = torch.tanh(ih[1] + f * hh[1])

        return c + f * (hx - c)


class IndRNNCellBase(nn.Module):
    """Base class for Independently RNN cells as presented in [Li2018]_.

    References
    ----------
    .. [Li2018] Li, S. et al. (2018) Independently Recurrent Neural Network (IndRNN): Building A Longer and Deeper RNN,
        Proceedings of the IEEE Computer Society Conference on Computer Vision and Pattern Recognition, (1), pp.
        5457–5466. doi: 10.1109/CVPR.2018.00572.
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        conv_dim: int,
        kernel_size: int,
        dilation: int,
        bias: bool = True,
    ):
        """Inits :class:`IndRNNCellBase`.

        Parameters
        ----------
        input_size : int
            Number of input channels.
        hidden_size : int
            Number of hidden channels.
        conv_dim : int
            Number of dimensions of the convolutional layer.
        kernel_size : int
            Size of the convolutional kernel.
        dilation : int
            Dilation of the convolutional kernel.
        bias : bool
            Whether to use bias. Default is ``True``.
        """
        super().__init__()

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.kernel_size = kernel_size
        self.bias = bias
        self.conv_dim = conv_dim
        self.conv_class = self.determine_conv_class(conv_dim)

        self.ih = self.conv_class(
            input_size,
            hidden_size,
            kernel_size,
            padding=torch.div(dilation * (kernel_size - 1), 2, rounding_mode="trunc").item(),
            dilation=dilation,
            bias=bias,
        )

        if self.conv_dim == 2:
            self.hh = nn.Parameter(
                nn.init.normal_(torch.empty(1, hidden_size, 1, 1), std=1.0 / (hidden_size * (1 + kernel_size**2)))
            )
        elif self.conv_dim == 3:
            self.hh = nn.Parameter(
                nn.init.normal_(torch.empty(1, hidden_size, 1, 1, 1), std=1.0 / (hidden_size * (1 + kernel_size**2)))
            )

        self.reset_parameters()

    def reset_parameters(self):
        """Reset the parameters."""
        self.ih.weight.data = self.orthotogonalize_weights(self.ih.weight.data)

        nn.init.normal_(self.ih.weight, std=1.0 / (self.hidden_size * (1 + self.kernel_size**2)))

        if self.bias is True:
            nn.init.zeros_(self.ih.bias)

    @staticmethod
    def orthotogonalize_weights(weights: torch.Tensor, chunks: int = 1) -> torch.Tensor:
        """Orthogonalize the weights."""
        return torch.cat([nn.init.orthogonal_(w) for w in weights.chunk(chunks, 0)], 0)

    @staticmethod
    def determine_conv_class(n_dim: int) -> nn.Module:
        """Determine the convolutional class."""
        if n_dim == 1:
            return nn.Conv1d
        if n_dim == 2:
            return nn.Conv2d
        if n_dim == 3:
            return nn.Conv3d
        raise NotImplementedError("No convolution of this dimensionality implemented")

    def extra_repr(self):
        """Extra information about the module, used for printing."""
        s = "{input_size}, {hidden_size}"
        if "bias" in self.__dict__ and self.bias is not True:
            s += ", bias={bias}"
        if "nonlinearity" in self.__dict__ and self.nonlinearity != "tanh":
            s += ", nonlinearity={nonlinearity}"
        return s.format(**self.__dict__)

    def check_forward_input(self, _input: torch.Tensor):
        """Check forward input."""
        if _input.size(1) != self.input_size:
            raise RuntimeError(f"input has inconsistent input_size: got {_input.size(1)}, expected {self.input_size}")

    def check_forward_hidden(self, _input, hx, hidden_label=""):
        """Check forward hidden."""
        if _input.size(0) != hx.size(0):
            raise RuntimeError(
                f"Input batch size {_input.size(0)} doesn't match hidden{hidden_label} batch size {hx.size(0)}"
            )

        if hx.size(1) != self.hidden_size:
            raise RuntimeError(
                f"hidden{hidden_label} has inconsistent hidden_size: got {hx.size(1)}, expected {self.hidden_size}"
            )


class IndRNNCell(IndRNNCellBase):
    """Base class for Independently RNN cells as presented in [Li2018]_.

    References
    ----------
    .. [Li2018] Li, S. et al. (2018) Independently Recurrent Neural Network (IndRNN): Building A Longer and Deeper RNN,
        Proceedings of the IEEE Computer Society Conference on Computer Vision and Pattern Recognition, (1), pp.
        5457–5466. doi: 10.1109/CVPR.2018.00572.
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        conv_dim: int,
        kernel_size: int,
        dilation: int,
        bias: bool = True,
    ):
        """Inits :class:`IndRNNCell`.

        Parameters
        ----------
        input_size : int
            Number of input channels.
        hidden_size : int
            Number of hidden channels.
        conv_dim : int
            Number of dimensions of the convolutional layer.
        kernel_size : int
            Size of the convolutional kernel.
        dilation : int
            Dilation of the convolutional kernel.
        bias : bool
            Whether to use bias. Default is ``True``.
        """
        super().__init__(input_size, hidden_size, conv_dim, kernel_size, dilation, bias)
        self.conv_dim = conv_dim

    def forward(self, _input: torch.Tensor, hx: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`IndRNNCell`."""
        if self.conv_dim == 3:
            # TODO: Check if this is correct
            _input = _input.unsqueeze(0)
            hx = hx.permute(1, 0, 2, 3).unsqueeze(0)

        return nn.ReLU()(self.ih(_input) + self.hh * hx)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/neural_types/elements.py

from abc import ABC, ABCMeta
from typing import Dict, Optional, Tuple

from atommic.core.neural_types.comparison import NeuralTypeComparisonResult

__all__ = ["ElementType", "LossType", "VoidType"]


class ElementType(ABC):
    """Abstract class defining semantics of the tensor elements. We are relying on Python for inheritance checking"""

    def __str__(self):
        """Override this method to provide a human readable representation of the type"""
        return self.__doc__

    def __repr__(self):
        """Override this method to provide a human readable representation of the type"""
        return self.__class__.__name__

    @property
    def type_parameters(self) -> Dict:
        """Override this property to parametrize your type. For example, you can specify 'storage' type such as float,
        int, bool with 'dtype' keyword. Another example, is if you want to represent a signal with a particular
        property (say, sample frequency), then you can put sample_freq->value in there. When two types are compared
        their type_parameters must match."
        """
        return {}

    @property
    def fields(self) -> Optional[Tuple]:
        """This should be used to logically represent tuples/structures. For example, if you want to represent a \
        bounding box (x, y, width, height) you can put a tuple with names ('x', y', 'w', 'h') in here. Under the \
        hood this should be converted to the last tensor dimension of fixed size = len(fields). When two types are \
        compared their fields must match."""
        return None

    def compare(self, second) -> NeuralTypeComparisonResult:
        """Override this method to provide a comparison between two types."""
        # First, check general compatibility
        first_t = type(self)
        second_t = type(second)

        if first_t == second_t:
            result = NeuralTypeComparisonResult.SAME
        elif issubclass(first_t, second_t):
            result = NeuralTypeComparisonResult.LESS
        elif issubclass(second_t, first_t):
            result = NeuralTypeComparisonResult.GREATER
        else:
            result = NeuralTypeComparisonResult.INCOMPATIBLE

        if result != NeuralTypeComparisonResult.SAME:
            return result
        # now check that all parameters match
        check_params = set(self.type_parameters.keys()) == set(second.type_parameters.keys())
        if not check_params:
            return NeuralTypeComparisonResult.SAME_TYPE_INCOMPATIBLE_PARAMS
        for k1, v1 in self.type_parameters.items():
            if v1 is None or second.type_parameters[k1] is None:
                # Treat None as Void
                continue
            if v1 != second.type_parameters[k1]:
                return NeuralTypeComparisonResult.SAME_TYPE_INCOMPATIBLE_PARAMS
                # check that all fields match
        if self.fields == second.fields:
            return NeuralTypeComparisonResult.SAME
        return NeuralTypeComparisonResult.INCOMPATIBLE


class VoidType(ElementType):
    """
    Void-like type which is compatible with everything. It is a good practice to use this type only as necessary.
    For example, when you need template-like functionality.
    """

    def compare(cls, second: ABCMeta) -> NeuralTypeComparisonResult:  # pylint: disable=arguments-renamed
        """Void type is compatible with everything."""
        return NeuralTypeComparisonResult.SAME


class LossType(ElementType):
    """Element type to represent outputs of Loss modules"""



# coding=utf-8
__author__ = "Dimitris Karkalousos"


import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.unet_base.unet_block import NormUnet
from atommic.core.classes.common import typecheck

__all__ = ["UNet"]


class UNet(BaseMRIReconstructionModel):
    """Implementation of the UNet, as presented in [Ronneberger2015]_.

    References
    ----------
    .. [Ronneberger2015] O. Ronneberger, P. Fischer, and Thomas Brox. U-net: Convolutional networks for biomedical
        image segmentation. In International Conference on Medical image computing and computer-assisted intervention,
        pages 234–241. Springer, 2015.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`UNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.in_channels = cfg_dict.get("in_channels", 2)
        self.reconstruction_module = NormUnet(
            chans=cfg_dict.get("channels"),
            num_pools=cfg_dict.get("pooling_layers"),
            in_chans=self.in_channels,
            out_chans=cfg_dict.get("out_channels", 2),
            padding_size=cfg_dict.get("padding_size", 11),
            drop_prob=cfg_dict.get("dropout", 0.0),
            normalize=cfg_dict.get("normalize", True),
            norm_groups=cfg_dict.get("norm_groups", 2),
        )

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,  # pylint: disable=unused-argument
        sensitivity_maps: torch.Tensor,  # pylint: disable=unused-argument
        mask: torch.Tensor,  # pylint: disable=unused-argument
        initial_prediction: torch.Tensor,
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`UNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        if self.in_channels == 1 and initial_prediction.shape[-1] == 2:
            initial_prediction = torch.abs(torch.view_as_complex(initial_prediction))
        prediction = self.reconstruction_module(initial_prediction.unsqueeze(self.coil_dim)).squeeze(self.coil_dim)
        if self.in_channels == 2:
            prediction = torch.view_as_complex(prediction)
        return prediction



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Any, List, Optional, Tuple, Union

import torch

from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import complex_conj, complex_mul
from atommic.collections.reconstruction.nn.rim_base.conv_layers import ConvNonlinear
from atommic.collections.reconstruction.nn.rim_base.rnn_cells import ConvGRUCell


class GRUConv2d(torch.nn.Module):
    """Implementation of a GRU followed by a number of 2D convolutions inspired by [Qin2019]_.

    References
    ----------
    .. [Qin2019] C. Qin, J. Schlemper, J. Caballero, A. N. Price, J. V. Hajnal and D. Rueckert, "Convolutional
        Recurrent Neural Networks for Dynamic MR Image Reconstruction," in IEEE Transactions on Medical Imaging, vol.
        38, no. 1, pp. 280-290, Jan. 2019, doi: 10.1109/TMI.2018.2863670.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        hidden_channels,
        n_convs=3,
        activation="ReLU",
        batchnorm=False,  # pylint: disable=unused-argument
    ):
        """Inits :class:`GRUConv2d`.

        Parameters
        ----------
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        hidden_channels : int
            Number of hidden channels.
        n_convs : int, optional
            Number of convolutional layers. Default is ``3``.
        activation : torch.nn.Module, optional
            Activation function. Default is ``nn.ReLU()``.
        batchnorm : bool, optional
            If True a batch normalization layer is applied after every convolution. Default is ``False``.
        """
        super().__init__()

        self.layers = torch.nn.ModuleList()
        self.layers.append(
            ConvGRUCell(
                in_channels,
                hidden_channels,
                conv_dim=2,
                kernel_size=3,
                dilation=1,
                bias=False,
            )
        )
        for _ in range(n_convs):
            self.layers.append(
                ConvNonlinear(
                    hidden_channels,
                    hidden_channels,
                    conv_dim=2,
                    kernel_size=3,
                    dilation=1,
                    bias=False,
                    nonlinear=activation,
                )
            )
        self.layers.append(
            torch.nn.Sequential(
                ConvNonlinear(
                    hidden_channels,
                    out_channels,
                    conv_dim=2,
                    kernel_size=3,
                    dilation=1,
                    bias=False,
                    nonlinear=activation,
                )
            )
        )

        self.hidden_channels = hidden_channels

    def forward(self, x, hx: Optional[torch.Tensor] = None):
        """Forward pass of :class:`GRUConv2d`.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.
        hx : torch.Tensor, optional
            Hidden state. Default is ``None``.

        Returns
        -------
        torch.Tensor
            Convoluted output.
        """
        if hx is None:
            hx = x.new_zeros((x.size(0), self.hidden_channels, *x.size()[2:]))

        for i, layer in enumerate(self.layers):
            x = layer(x, hx) if i == 0 else layer(x)
        return x


class DataConsistencyLayer(torch.nn.Module):
    """Data consistency layer for the CRNN, inspired by [Qin2019]_.

    References
    ----------
    .. [Qin2019] C. Qin, J. Schlemper, J. Caballero, A. N. Price, J. V. Hajnal and D. Rueckert, "Convolutional
        Recurrent Neural Networks for Dynamic MR Image Reconstruction," in IEEE Transactions on Medical Imaging, vol.
        38, no. 1, pp. 280-290, Jan. 2019, doi: 10.1109/TMI.2018.2863670.
    """

    def __init__(self):
        """Initializes the data consistency layer."""
        super().__init__()
        self.dc_weight = torch.nn.Parameter(torch.ones(1))

    def forward(self, pred_kspace: torch.Tensor, ref_kspace: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
        """Forward pass of :class:`DataConsistencyLayer`.

        Parameters
        ----------
        pred_kspace : torch.Tensor
            Predicted k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        ref_kspace : torch.Tensor
            Reference k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        """
        zero = torch.zeros(1, 1, 1, 1, 1).to(pred_kspace)
        return torch.where(mask.bool(), pred_kspace - ref_kspace, zero) * self.dc_weight


class RecurrentConvolutionalNetBlock(torch.nn.Module):
    """Model block for Recurrent Convolution Neural Network inspired by [Qin2019]_.

    References
    ----------
    .. [Qin2019] C. Qin, J. Schlemper, J. Caballero, A. N. Price, J. V. Hajnal and D. Rueckert, "Convolutional
        Recurrent Neural Networks for Dynamic MR Image Reconstruction," in IEEE Transactions on Medical Imaging, vol.
        38, no. 1, pp. 280-290, Jan. 2019, doi: 10.1109/TMI.2018.2863670.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        num_iterations: int = 10,
        fft_centered: bool = False,
        fft_normalization: str = "backward",
        spatial_dims: Optional[Tuple[int, int]] = None,
        coil_dim: int = 1,
        no_dc: bool = False,
    ):
        """Inits :class:`RecurrentConvolutionalNetBlock`.

        Parameters
        ----------
        model : torch.nn.Module
            Model to apply soft data consistency.
        num_iterations : int, optional
            Number of iterations. Default is ``10``.
        fft_centered : bool, optional
            Whether to use centered FFT. Default is ``False``.
        fft_normalization : str, optional
            Whether to use normalized FFT. Default is ``"backward"``.
        spatial_dims : tuple, optional
            Spatial dimensions of the input. Default is ``None``.
        coil_dim : int, optional
            Dimension of the coil. Default is ``1``.
        no_dc : bool, optional
            Whether to remove the DC component. Default is ``False``.
        """
        super().__init__()

        self.model = model
        self.num_iterations = num_iterations
        self.fft_centered = fft_centered
        self.fft_normalization = fft_normalization
        self.spatial_dims = spatial_dims if spatial_dims is not None else [-2, -1]
        self.coil_dim = coil_dim
        self.no_dc = no_dc

        self.dc_weight = torch.nn.Parameter(torch.ones(1))

    def sens_expand(self, x: torch.Tensor, sens_maps: torch.Tensor) -> torch.Tensor:
        """Combines the sensitivity maps with coil-combined data to get multicoil data.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        sens_maps : torch.Tensor
            Coil Sensitivity maps.

        Returns
        -------
        torch.Tensor
            Expanded multicoil data.
        """
        return fft2(
            complex_mul(x, sens_maps),
            centered=self.fft_centered,
            normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
        )

    def sens_reduce(self, x: torch.Tensor, sens_maps: torch.Tensor) -> torch.Tensor:
        """Combines the sensitivity maps with multicoil data to get coil-combined data.

        Parameters
        ----------
        x : torch.Tensor
            Input data.
        sens_maps : torch.Tensor
            Coil Sensitivity maps.

        Returns
        -------
        torch.Tensor
            SENSE coil-combined reconstruction.
        """
        x = ifft2(x, centered=self.fft_centered, normalization=self.fft_normalization, spatial_dims=self.spatial_dims)
        return complex_mul(x, complex_conj(sens_maps)).sum(dim=self.coil_dim)

    def forward(
        self,
        ref_kspace: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
    ) -> List[Union[torch.Tensor, Any]]:
        """Forward pass of :class:`RecurrentConvolutionalNetBlock`.

        Parameters
        ----------
        ref_kspace : torch.Tensor
            Reference k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]

        Returns
        -------
        torch.Tensor
            Reconstructed image. Shape [batch_size, n_x, n_y, 2]
        """
        zero = torch.zeros(1, 1, 1, 1, 1).to(ref_kspace)
        pred = ref_kspace.clone()

        preds = []
        for _ in range(self.num_iterations):
            soft_dc = torch.where(mask.bool(), pred - ref_kspace, zero) * self.dc_weight

            prediction = self.sens_reduce(pred, sensitivity_maps)
            prediction = self.model(prediction.permute(0, 3, 1, 2)).permute(0, 2, 3, 1) + prediction
            prediction = self.sens_expand(prediction.unsqueeze(self.coil_dim), sensitivity_maps)

            if not self.no_dc:
                prediction = pred - soft_dc - prediction

            pred = prediction

            preds.append(prediction)

        return preds



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/config/schedulers.py

from dataclasses import dataclass
from functools import partial
from typing import Any, Dict, Optional


@dataclass
class SchedulerParams:
    """Base configuration for all schedulers."""

    last_epoch: int = -1


@dataclass
class SquareRootConstantSchedulerParams(SchedulerParams):
    """Base configuration for all schedulers. It is not derived from Config as it is not a atommic object (and in
    particular it doesn't need a name).
    """

    constant_steps: Optional[float] = None
    constant_ratio: Optional[float] = None


@dataclass
class WarmupSchedulerParams(SchedulerParams):
    """Base configuration for all schedulers."""

    max_steps: int = 0
    warmup_steps: Optional[float] = None
    warmup_ratio: Optional[float] = None


@dataclass
class WarmupHoldSchedulerParams(WarmupSchedulerParams):
    """Base configuration for all schedulers."""

    hold_steps: Optional[float] = None
    hold_ratio: Optional[float] = None
    min_lr: float = 0.0


@dataclass
class WarmupAnnealingHoldSchedulerParams(WarmupSchedulerParams):
    """Base configuration for all schedulers."""

    constant_steps: Optional[float] = None
    constant_ratio: Optional[float] = None
    min_lr: float = 0.0


@dataclass
class SquareAnnealingParams(WarmupSchedulerParams):
    """Square Annealing parameter config"""

    min_lr: float = 1e-5


@dataclass
class SquareRootAnnealingParams(WarmupSchedulerParams):
    """Square Root Annealing parameter config"""

    min_lr: float = 0.0


@dataclass
class CosineAnnealingParams(WarmupAnnealingHoldSchedulerParams):
    """Cosine Annealing parameter config"""

    min_lr: float = 0.0


@dataclass
class NoamAnnealingParams(WarmupSchedulerParams):
    """Cosine Annealing parameter config"""

    min_lr: float = 0.0


@dataclass
class NoamHoldAnnealingParams(WarmupHoldSchedulerParams):
    """Polynomial Hold Decay Annealing parameter config. It is not derived from Config as it is not a atommic object
    (and in particular it doesn't need a name).
    """

    decay_rate: float = 0.5


@dataclass
class WarmupAnnealingParams(WarmupSchedulerParams):
    """Warmup Annealing parameter config"""

    warmup_ratio: Optional[float] = None


@dataclass
class InverseSquareRootAnnealingParams(WarmupSchedulerParams):
    """Inverse Square Root Annealing parameter config"""


@dataclass
class PolynomialDecayAnnealingParams(WarmupSchedulerParams):
    """Polynomial Decay Annealing parameter config"""

    power: float = 1.0
    cycle: bool = False


@dataclass
class PolynomialHoldDecayAnnealingParams(WarmupSchedulerParams):
    """Polynomial Hold Decay Annealing parameter config"""

    power: float = 1.0
    cycle: bool = False


@dataclass
class StepLRParams(SchedulerParams):
    """Config for StepLR."""

    step_size: float = 0.1
    gamma: float = 0.1


@dataclass
class ExponentialLRParams(SchedulerParams):
    """Config for ExponentialLR."""

    gamma: float = 0.9


@dataclass
class ReduceLROnPlateauParams:
    """Config for ReduceLROnPlateau."""

    mode: str = "min"
    factor: float = 0.1
    patience: int = 10
    verbose: bool = False
    threshold: float = 1e-4
    threshold_mode: str = "rel"
    cooldown: int = 0
    min_lr: float = 0
    eps: float = 1e-8


@dataclass
class CyclicLRParams(SchedulerParams):
    """Config for CyclicLR."""

    base_lr: float = 0.001
    max_lr: float = 0.1
    step_size_up: int = 2000
    step_size_down: Optional[int] = None
    mode: str = "triangular"
    gamma: float = 1.0
    scale_mode: str = "cycle"
    # scale_fn is not supported
    cycle_momentum: bool = True
    base_momentum: float = 0.8
    max_momentum: float = 0.9


def register_scheduler_params(name: str, scheduler_params: SchedulerParams):
    """Checks if the scheduler config name exists in the registry, and if it doesn't, adds it. This allows custom
    schedulers to be added and called by name during instantiation.

    Parameters
    ----------
    name : str
        Name of the scheduler. Will be used as key to retrieve the scheduler.
    scheduler_params : SchedulerParams
        SchedulerParams class to be added to the registry.
    """
    if name in AVAILABLE_SCHEDULER_PARAMS:
        raise ValueError(f"Cannot override pre-existing optimizers. Conflicting optimizer name = {name}")

    AVAILABLE_SCHEDULER_PARAMS[name] = scheduler_params  # type: ignore


def get_scheduler_config(name: str, **kwargs: Optional[Dict[str, Any]]) -> partial:
    """Convenience method to obtain a SchedulerParams class and partially instantiate it with optimizer kwargs.

    Parameters
    ----------
    name : str
        Name of the SchedulerParams. in the registry..
    kwargs : Dict[str, Any], optional
        Optional kwargs of the optimizer used during instantiation.

    Returns
    -------
    A partially instantiated SchedulerParams.
    """
    if name not in AVAILABLE_SCHEDULER_PARAMS:
        raise ValueError(
            f"Cannot resolve scheduler parameters '{name}'. Available scheduler parameters are : "
            f"{AVAILABLE_SCHEDULER_PARAMS.keys()}"
        )

    return partial(AVAILABLE_SCHEDULER_PARAMS[name], **kwargs)


AVAILABLE_SCHEDULER_PARAMS = {
    "SchedulerParams": SchedulerParams,
    "WarmupPolicyParams": WarmupSchedulerParams,
    "WarmupHoldPolicyParams": WarmupHoldSchedulerParams,
    "WarmupAnnealingHoldSchedulerParams": WarmupAnnealingHoldSchedulerParams,
    "SquareAnnealingParams": SquareAnnealingParams,
    "SquareRootAnnealingParams": SquareRootAnnealingParams,
    "InverseSquareRootAnnealingParams": InverseSquareRootAnnealingParams,
    "SquareRootConstantSchedulerParams": SquareRootConstantSchedulerParams,
    "CosineAnnealingParams": CosineAnnealingParams,
    "NoamAnnealingParams": NoamAnnealingParams,
    "NoamHoldAnnealingParams": NoamHoldAnnealingParams,
    "WarmupAnnealingParams": WarmupAnnealingParams,
    "PolynomialDecayAnnealingParams": PolynomialDecayAnnealingParams,
    "PolynomialHoldDecayAnnealingParams": PolynomialHoldDecayAnnealingParams,
    "ReduceLROnPlateauParams": ReduceLROnPlateauParams,
}



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/blob/main/nemo/core/config/modelPT.py

from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from omegaconf import MISSING

from atommic.core.classes.dataset import DatasetConfig
from atommic.core.conf.optimizers import OptimizerParams
from atommic.core.conf.schedulers import SchedulerParams
from atommic.core.conf.trainer import TrainerConfig
from atommic.utils.exp_manager import ExpManagerConfig


@dataclass
class SchedConfig:
    """Configuration for the scheduler."""

    name: str = MISSING
    min_lr: float = 0.0
    last_epoch: int = -1


@dataclass
class OptimConfig:
    """Configuration for the optimizer."""

    name: str = MISSING
    sched: Optional[SchedConfig] = None


@dataclass
class ModelConfig:
    """Configuration for the model."""

    train_ds: Optional[DatasetConfig] = None
    validation_ds: Optional[DatasetConfig] = None
    test_ds: Optional[DatasetConfig] = None
    optim: Optional[OptimConfig] = None


@dataclass
class HydraConfig:
    """Configuration for the hydra framework."""

    run: Dict[str, Any] = field(default_factory=lambda: {"dir": "."})
    job_logging: Dict[str, Any] = field(default_factory=lambda: {"root": {"handlers": None}})


@dataclass
class atommicConfig:
    """Configuration for the atommic framework."""

    name: str = MISSING
    model: ModelConfig = MISSING
    trainer: TrainerConfig = TrainerConfig(
        strategy="ddp",
        enable_checkpointing=False,
        logger=False,
        log_every_n_steps=1,
        accelerator="gpu",
    )
    exp_manager: Optional[Any] = ExpManagerConfig()
    hydra: HydraConfig = HydraConfig()


class ModelConfigBuilder:
    """Builder for the ModelConfig class."""

    def __init__(self, model_cfg: ModelConfig):
        """Inits :class:`ModelConfigBuilder`.

        A Model Config Builder is a utility class that accepts a ModelConfig dataclass, and via a set of utility
        methods (that are implemented by the subclassed ModelConfigBuilder), builds a finalized ModelConfig that can be
         supplied to a atommicModel dataclass as the `model` component.

        Subclasses *must* implement the private method `_finalize_cfg`.
            Inside this method, they must update `self.model_cfg` with all interdependent config
            options that need to be set (either updated by user explicitly or with their default value).
            The updated model config must then be preserved in `self.model_cfg`.

        Example
        -------
        # Create the config builder
        config_builder = <subclass>ModelConfigBuilder()
        # Update the components of the config that are modifiable
        config_builder.set_X(X)
        config_builder.set_Y(Y)
        # Create a "finalized" config dataclass that will contain all the updates
        # that were specified by the builder
        model_config = config_builder.build()
        # Use model config as is (or further update values), then create a new Model
        model = atommic.<domain>.models.<ModelName>Model(cfg=model_config, trainer=Trainer())
        Supported build methods:
        -   set_train_ds: All model configs can accept a subclass of `DatasetConfig` as their
                training conf. Subclasses can override this method to enable auto-complete
                by replacing `Optional[DatasetConfig]` with `Optional[<subclass of DatasetConfig>]`.
        -   set_validation_ds: All model configs can accept a subclass of `DatasetConfig` as their
                validation conf. Subclasses can override this method to enable auto-complete
                by replacing `Optional[DatasetConfig]` with `Optional[<subclass of DatasetConfig>]`.
        -   set_test_ds: All model configs can accept a subclass of `DatasetConfig` as their
                test conf. Subclasses can override this method to enable auto-complete
                by replacing `Optional[DatasetConfig]` with `Optional[<subclass of DatasetConfig>]`.
        -   set_optim: A build method that supports changes to the Optimizer (and optionally,
                the Scheduler) used for training the model. The function accepts two inputs -
                `cfg`: A subclass of `OptimizerParams` - any OptimizerParams subclass can be used,
                    in order to select an appropriate Optimizer. Examples: AdamParams.
                `sched_cfg`: A subclass of `SchedulerParams` - any SchedulerParams subclass can be used,
                    in order to select an appropriate Scheduler. Examples: CosineAnnealingParams.
                    Note that this argument is optional.
        -   build(): The method which should return a "finalized" ModelConfig dataclass.
                Subclasses *should* always override this method, and update the signature
                of this method with the return type of the Dataclass, so that it enables
                autocomplete for the user.
                Example:
                    def build(self) -> EncDecCTCConfig:
                        return super().build()
        Any additional build methods must be added by subclasses of ModelConfigBuilder.

        Parameters
        ----------
        model_cfg : ModelConfig
            The model config dataclass to be updated.

        Returns
        -------
        The updated model config dataclass.
        """
        self.model_cfg = model_cfg
        self.train_ds_cfg = None
        self.validation_ds_cfg = None
        self.test_ds_cfg = None
        self.optim_cfg = None

    def set_train_ds(self, cfg: Optional[DatasetConfig] = None):
        """Set the training dataset configuration."""
        self.model_cfg.train_ds = cfg

    def set_validation_ds(self, cfg: Optional[DatasetConfig] = None):
        """Set the validation dataset configuration."""
        self.model_cfg.validation_ds = cfg

    def set_test_ds(self, cfg: Optional[DatasetConfig] = None):
        """Set the test dataset configuration."""
        self.model_cfg.test_ds = cfg

    def set_optim(self, cfg: OptimizerParams, sched_cfg: Optional[SchedulerParams] = None):
        """Set the optimizer configuration."""

        @dataclass
        class WrappedOptimConfig(OptimConfig, cfg.__class__):  # type: ignore
            """A wrapper class for the OptimizerParams dataclass."""

        # Setup optim
        optim_name = cfg.__class__.__name__.replace("Params", "").lower()
        wrapped_cfg = WrappedOptimConfig(name=optim_name, sched=None, **vars(cfg))

        if sched_cfg is not None:

            @dataclass
            class WrappedSchedConfig(SchedConfig, sched_cfg.__class__):  # type: ignore
                """A wrapper class for the SchedulerParams dataclass."""

            # Setup scheduler
            sched_name = sched_cfg.__class__.__name__.replace("Params", "")
            wrapped_sched_cfg = WrappedSchedConfig(name=sched_name, **vars(sched_cfg))

            wrapped_cfg.sched = wrapped_sched_cfg

        self.model_cfg.optim = wrapped_cfg

    def _finalize_cfg(self):
        """Finalize the model configuration."""
        raise NotImplementedError()

    def build(self) -> ModelConfig:
        """Validate config"""
        self._finalize_cfg()

        return self.model_cfg



# coding=utf-8
__author__ = "Dimitris Karkalousos"


import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning import Trainer

from atommic.collections.common.nn.base import BaseSensitivityModel
from atommic.collections.common.parts.fft import fft2, ifft2
from atommic.collections.common.parts.utils import (
    check_stacked_complex,
    coil_combination_method,
    complex_conj,
    complex_mul,
)
from atommic.collections.reconstruction.nn.base import BaseMRIReconstructionModel
from atommic.collections.reconstruction.nn.unet_base.unet_block import NormUnet
from atommic.core.classes.common import typecheck

__all__ = ["JointICNet"]


class JointICNet(BaseMRIReconstructionModel):
    """Implementation of the Joint Deep Model-Based MR Image and Coil Sensitivity Reconstruction Network (Joint-ICNet),
     as presented in [Jun2021]_.

    References
    ----------
    .. [Jun2021] Jun, Yohan, et al. “Joint Deep Model-Based MR Image and Coil Sensitivity Reconstruction Network
        (Joint-ICNet) for Fast MRI.” 2021 IEEE/CVF Conference on Computer Vision and Pattern Recognition (CVPR), IEEE,
        2021, pp. 5266–75. DOI.org (Crossref), https://doi.org/10.1109/CVPR46437.2021.00523.

    """

    def __init__(self, cfg: DictConfig, trainer: Trainer = None):
        """Inits :class:`JointICNet`.

        Parameters
        ----------
        cfg : DictConfig
            Configuration.
        trainer : Trainer, optional
            PyTorch Lightning trainer. Default is ``None``.
        """
        super().__init__(cfg=cfg, trainer=trainer)

        cfg_dict = OmegaConf.to_container(cfg, resolve=True)

        self.num_iter = cfg_dict.get("num_iter")

        self.kspace_model = NormUnet(
            cfg_dict.get("kspace_unet_num_filters"),
            cfg_dict.get("kspace_unet_num_pool_layers"),
            in_chans=2,
            out_chans=2,
            drop_prob=cfg_dict.get("kspace_unet_dropout_probability"),
            padding_size=cfg_dict.get("kspace_unet_padding_size"),
            normalize=cfg_dict.get("kspace_unet_normalize"),
        )

        self.image_model = NormUnet(
            cfg_dict.get("imspace_unet_num_filters"),
            cfg_dict.get("imspace_unet_num_pool_layers"),
            in_chans=2,
            out_chans=2,
            drop_prob=cfg_dict.get("imspace_unet_dropout_probability"),
            padding_size=cfg_dict.get("imspace_unet_padding_size"),
            normalize=cfg_dict.get("imspace_unet_normalize"),
        )

        self.sens_net = BaseSensitivityModel(
            cfg_dict.get("sens_unet_num_filters"),
            cfg_dict.get("sens_unet_num_pool_layers"),
            mask_center=cfg_dict.get("sens_unet_mask_center"),
            fft_centered=self.fft_centered,
            fft_normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
            coil_dim=self.coil_dim,
            mask_type=cfg_dict.get("coil_sensitivity_maps_nn_mask_type", "2D"),
            drop_prob=cfg_dict.get("sens_unet_dropout_probability"),
            padding_size=cfg_dict.get("sens_unet_padding_size"),
            normalize=cfg_dict.get("sens_unet_normalize"),
        )

        self.conv_out = torch.nn.Conv2d(in_channels=2, out_channels=2, kernel_size=1)

        self.reg_param_I = torch.nn.Parameter(torch.ones(self.num_iter))
        self.reg_param_F = torch.nn.Parameter(torch.ones(self.num_iter))
        self.reg_param_C = torch.nn.Parameter(torch.ones(self.num_iter))

        self.lr_image = torch.nn.Parameter(torch.ones(self.num_iter))
        self.lr_sens = torch.nn.Parameter(torch.ones(self.num_iter))

    def update_C(
        self,
        idx: int,
        DC_sens: torch.Tensor,
        image: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        y: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        r"""Updates the coil sensitivity maps.

        .. math::
            C = (1 - 2 * '\'lambda_{k}^{C} * ni_{k}) * C_{k}

            C = 2 * '\'lambda_{k}^{C} * ni_{k} * D_{C}(F^-1(b))

            A(x_{k}) = M * F * (C * x_{k})

            C = 2 * ni_{k} * F^-1(M.T * (M * F * (C * x_{k}) - b)) * x_{k}^*

        Parameters
        ----------
        idx : int
            The current iteration index.
        DC_sens : torch.Tensor
            The initial coil sensitivity maps. Shape [batch_size, num_coils, num_sens_maps, num_rows, num_cols].
        image : torch.Tensor
            The predicted image. Shape [batch_size, num_coils, num_rows, num_cols].
        sensitivity_maps : torch.Tensor
            The coil sensitivity maps. Shape [batch_size, num_coils, num_sens_maps, num_rows, num_cols].
        y : torch.Tensor
            The subsampled k-space data. Shape [batch_size, num_coils, num_rows, num_cols].
        mask : torch.Tensor
            The subsampled mask. Shape [batch_size, 1, num_rows, num_cols].

        Returns
        -------
        sensitivity_maps : torch.Tensor
            The updated coil sensitivity maps. Shape [batch_size, num_coils, num_sens_maps, num_rows, num_cols].
        """
        # (1 - 2 * lambda_{k}^{C} * ni_{k}) * C_{k}
        sense_term_1 = (1 - 2 * self.reg_param_C[idx] * self.lr_sens[idx]) * sensitivity_maps
        # 2 * lambda_{k}^{C} * ni_{k} * D_{C}(F^-1(b))
        sense_term_2 = 2 * self.reg_param_C[idx] * self.lr_sens[idx] * DC_sens
        # A(x_{k}) = M * F * (C * x_{k})
        sense_term_3_A = fft2(
            complex_mul(image.unsqueeze(self.coil_dim), sensitivity_maps),
            centered=self.fft_centered,
            normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
        )
        sense_term_3_A = torch.where(mask == 0, torch.tensor([0.0], dtype=y.dtype).to(y.device), sense_term_3_A)
        # 2 * ni_{k} * F^-1(M.T * (M * F * (C * x_{k}) - b)) * x_{k}^*
        sense_term_3_mask = torch.where(
            mask == 1,
            torch.tensor([0.0], dtype=y.dtype).to(y.device),
            sense_term_3_A - y,
        )

        sense_term_3_backward = ifft2(
            sense_term_3_mask,
            centered=self.fft_centered,
            normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
        )
        sense_term_3 = 2 * self.lr_sens[idx] * sense_term_3_backward * complex_conj(image).unsqueeze(self.coil_dim)
        sensitivity_maps = sense_term_1 + sense_term_2 - sense_term_3
        return sensitivity_maps

    def update_X(
        self,
        idx: int,
        image: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        y: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        r"""Updates the image.

        .. math::
            x_{k} = (1 - 2 * '\'lamdba_{{k}_{I}} * mi_{k} - 2 * '\'lamdba_{{k}_{F}} * mi_{k}) * x_{k}

            x_{k} = 2 * mi_{k} * ('\'lambda_{{k}_{I}} * D_I(x_{k}) + '\'lambda_{{k}_{F}} * F^-1(D_F(f)))

            A(x{k} - b) = M * F * (C * x{k}) - b

            x_{k} = 2 * mi_{k} * A^* * (A(x{k} - b))

        Parameters
        ----------
        idx : int
            The current iteration index.
        image : torch.Tensor
            The predicted image. Shape [batch_size, num_coils, num_rows, num_cols].
        sensitivity_maps : torch.Tensor
            The coil sensitivity maps. Shape [batch_size, num_coils, num_sens_maps, num_rows, num_cols].
        y : torch.Tensor
            The subsampled k-space data. Shape [batch_size, num_coils, num_rows, num_cols].
        mask : torch.Tensor
            The subsampling mask. Shape [batch_size, 1, num_rows, num_cols].

        Returns
        -------
        image : torch.Tensor
            The updated image. Shape [batch_size, num_coils, num_rows, num_cols].
        """
        # (1 - 2 * lamdba_{k}_{I} * mi_{k} - 2 * lamdba_{k}_{F} * mi_{k}) * x_{k}
        image_term_1 = (
            1 - 2 * self.reg_param_I[idx] * self.lr_image[idx] - 2 * self.reg_param_F[idx] * self.lr_image[idx]
        ) * image
        # D_I(x_{k})
        image_term_2_DI = self.image_model(image.unsqueeze(self.coil_dim)).squeeze(self.coil_dim).contiguous()
        image_term_2_DF = ifft2(
            self.kspace_model(
                fft2(
                    image,
                    centered=self.fft_centered,
                    normalization=self.fft_normalization,
                    spatial_dims=self.spatial_dims,
                ).unsqueeze(self.coil_dim)
            )
            .squeeze(self.coil_dim)
            .contiguous(),
            centered=self.fft_centered,
            normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
        )
        # 2 * mi_{k} * (lambda_{k}_{I} * D_I(x_{k}) + lambda_{k}_{F} * F^-1(D_F(f)))
        image_term_2 = (
            2
            * self.lr_image[idx]
            * (self.reg_param_I[idx] * image_term_2_DI + self.reg_param_F[idx] * image_term_2_DF)
        )
        # A(x{k}) - b) = M * F * (C * x{k}) - b
        image_term_3_A = fft2(
            complex_mul(image.unsqueeze(self.coil_dim), sensitivity_maps),
            centered=self.fft_centered,
            normalization=self.fft_normalization,
            spatial_dims=self.spatial_dims,
        )
        image_term_3_A = torch.where(mask == 0, torch.tensor([0.0], dtype=y.dtype).to(y.device), image_term_3_A) - y
        # 2 * mi_{k} * A^* * (A(x{k}) - b))
        image_term_3_Aconj = complex_mul(
            ifft2(
                image_term_3_A,
                centered=self.fft_centered,
                normalization=self.fft_normalization,
                spatial_dims=self.spatial_dims,
            ),
            complex_conj(sensitivity_maps),
        ).sum(self.coil_dim)
        image_term_3 = 2 * self.lr_image[idx] * image_term_3_Aconj
        image = image_term_1 + image_term_2 - image_term_3
        return image

    # pylint: disable=arguments-differ
    @typecheck()
    def forward(
        self,
        y: torch.Tensor,
        sensitivity_maps: torch.Tensor,
        mask: torch.Tensor,
        initial_prediction: torch.Tensor,  # pylint: disable=unused-argument
        sigma: float = 1.0,  # pylint: disable=unused-argument
    ) -> torch.Tensor:
        """Forward pass of :class:`JointICNet`.

        Parameters
        ----------
        y : torch.Tensor
            Subsampled k-space data. Shape [batch_size, n_coils, n_x, n_y, 2]
        sensitivity_maps : torch.Tensor
            Coil sensitivity maps. Shape [batch_size, n_coils, n_x, n_y, 2]
        mask : torch.Tensor
            Subsampling mask. Shape [1, 1, n_x, n_y, 1]
        initial_prediction : torch.Tensor
            Initial prediction. Shape [batch_size, n_x, n_y, 2]
        sigma : float
            Noise level. Default is ``1.0``.

        Returns
        -------
        torch.Tensor
            Prediction of the final cascade. Shape [batch_size, n_x, n_y]
        """
        DC_sens = self.sens_net(y, mask, sensitivity_maps)
        sensitivity_maps = DC_sens.clone()
        image = coil_combination_method(
            ifft2(y, self.fft_centered, self.fft_normalization, self.spatial_dims),
            sensitivity_maps,
            self.coil_combination_method,
            self.coil_dim,
        )
        for idx in range(self.num_iter):
            sensitivity_maps = self.update_C(idx, DC_sens, image, sensitivity_maps, y, mask)
            image = self.update_X(idx, image, sensitivity_maps, y, mask)
        return check_stacked_complex(image)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from: https://github.com/NVIDIA/NeMo/tree/main/nemo/utils/cast_utils.py

import torch


def cast_tensor(x, from_dtype=torch.float16, to_dtype=torch.float32):
    """Cast a tensor from one dtype to another if it is of the specified dtype."""
    return x.to(dtype=to_dtype) if x.dtype == from_dtype else x


# pylint: disable=inconsistent-return-statements
def cast_all(x, from_dtype=torch.float16, to_dtype=torch.float32):
    """Cast all tensors in a dict or tuple from one dtype to another if they are of the specified dtype."""
    if isinstance(x, torch.Tensor):
        return cast_tensor(x, from_dtype=from_dtype, to_dtype=to_dtype)
    if isinstance(x, dict):
        new_dict = {}
        for k in x.keys():
            new_dict[k] = cast_all(x[k], from_dtype=from_dtype, to_dtype=to_dtype)
        return new_dict
    if isinstance(x, tuple):
        return tuple(cast_all(y, from_dtype=from_dtype, to_dtype=to_dtype) for y in x)



# coding=utf-8
__author__ = "Dimitris Karkalousos"

# Taken and adapted from:
# https://github.com/khammernik/sigmanet/blob/master/reconstruction/common/mytorch/models/sn.py

import numpy as np
import torch


def matrix_invert(xx, xy, yx, yy):
    """Invert a 2x2 matrix."""
    det = xx * yy - xy * yx
    return yy.div(det), -xy.div(det), -yx.div(det), xx.div(det)


class ComplexInstanceNorm(torch.nn.Module):
    """Complex instance normalization layer, as presented in [Chiheb2017]_.

    References
    ----------
    .. [Chiheb2017] Deep complex networks, T Chiheb, O Bilaniuk, D Serdyuk - … Conference on Learning Representations,
        2017, https://arxiv.org/abs/1705.09792
    """

    def __init__(self):
        """Inits :class:`ComplexInstanceNorm`."""
        super().__init__()
        self.mean = 0
        self.cov_xx_half = 1 / np.sqrt(2)
        self.cov_xy_half = 0
        self.cov_yx_half = 0
        self.cov_yy_half = 1 / np.sqrt(2)

    def complex_instance_norm(self, x):
        """Operates on images x of size [nBatch, nSmaps, nFE, nPE, 2]"""
        x_combined = torch.sum(x, dim=1, keepdim=True)
        mean = x_combined.mean(dim=(1, 2, 3), keepdim=True)
        x_m = x - mean
        self.mean = mean
        self.complex_pseudocovariance(x_m)

    def complex_pseudocovariance(self, data):
        """Data variable hast to be already mean-free! Operates on images x of size [nBatch, nSmaps, nFE, nPE, 2]"""
        if data.size(-1) != 2:
            raise AssertionError
        shape = data.shape

        # compute number of elements
        N = shape[2] * shape[3]

        # separate real/imaginary channel
        re, im = torch.unbind(data, dim=-1)

        # dimensions is now length of original shape - 1 (because channels are separated)
        dim = list(range(1, len(shape) - 1))

        # compute covariance entries. cxy = cyx
        cxx = (re * re).sum(dim=dim, keepdim=True) / (N - 1)
        cyy = (im * im).sum(dim=dim, keepdim=True) / (N - 1)
        cxy = (re * im).sum(dim=dim, keepdim=True) / (N - 1)

        # Eigenvalue decomposition C = V*S*inv(V)
        s1 = (cxx + cyy) / 2 - torch.sqrt((cxx + cyy) ** 2 / 4 - cxx * cyy + cxy**2)
        s2 = (cxx + cyy) / 2 + torch.sqrt((cxx + cyy) ** 2 / 4 - cxx * cyy + cxy**2)

        # compute eigenvectors
        v1x = s1 - cyy
        v1y = cxy
        v2x = s2 - cyy
        v2y = cxy

        # normalize eigenvectors
        norm1 = torch.sqrt(torch.sum(v1x * v1x + v1y * v1y, dim=dim, keepdim=True))
        norm2 = torch.sqrt(torch.sum(v2x * v2x + v2y * v2y, dim=dim, keepdim=True))

        v1x = v1x.div(norm1)
        v1y = v1y.div(norm1)

        v2x = v2x.div(norm2)
        v2y = v2y.div(norm2)

        # now we need the sqrt of the covariance matrix.
        # C^{-0.5} = V * sqrt(S) * inv(V)
        det = v1x * v2y - v2x * v1y
        s1 = torch.sqrt(s1).div(det)
        s2 = torch.sqrt(s2).div(det)

        self.cov_xx_half = v1x * v2y * s1 - v1y * v2x * s2
        self.cov_yy_half = v1x * v2y * s2 - v1y * v2x * s1
        self.cov_xy_half = v1x * v2x * (s2 - s1)
        self.cov_yx_half = v1y * v2y * (s1 - s2)

    def forward(self, _input):
        """Forward pass of :class:`ComplexInstanceNorm`."""
        return self.normalize(_input)

    def set_normalization(self, _input):
        """Set the normalization parameters for a given _input."""
        mean = torch.tensor([torch.mean(_input).item()]).to(_input)
        self.complex_pseudocovariance(_input - mean)
        self.mean = mean.unsqueeze(1).unsqueeze(1).unsqueeze(1)
        self.cov_xx_half = self.cov_xx_half.view(-1, 1, 1, 1)
        self.cov_xy_half = self.cov_xy_half.view(-1, 1, 1, 1)
        self.cov_yx_half = self.cov_yx_half.view(-1, 1, 1, 1)
        self.cov_yy_half = self.cov_yy_half.view(-1, 1, 1, 1)

    def normalize(self, x):
        """Normalize the _input x."""
        x_m = x - self.mean
        re, im = torch.unbind(x_m, dim=-1)

        cov_xx_half_inv, cov_xy_half_inv, cov_yx_half_inv, cov_yy_half_inv = matrix_invert(
            self.cov_xx_half, self.cov_xy_half, self.cov_yx_half, self.cov_yy_half
        )
        x_norm_re = cov_xx_half_inv * re + cov_xy_half_inv * im
        x_norm_im = cov_yx_half_inv * re + cov_yy_half_inv * im
        img = torch.stack([x_norm_re, x_norm_im], dim=-1)
        # img = img.clamp(-6, 6)
        return img

    def unnormalize(self, x):
        """Unnormalize the _input x."""
        re, im = torch.unbind(x, dim=-1)
        x_unnorm_re = self.cov_xx_half * re + self.cov_xy_half * im
        x_unnorm_im = self.cov_yx_half * re + self.cov_yy_half * im
        return torch.stack([x_unnorm_re, x_unnorm_im], dim=-1) + self.mean


class ComplexNormWrapper(torch.nn.Module):
    """Wrapper for complex normalization."""

    def __init__(self, model):
        """Inits :class:`ComplexNormWrapper`.

        Parameters
        ----------
        model : torch.nn.Module
            Model to be wrapped.
        """
        super().__init__()
        self.model = model
        self.complex_instance_norm = ComplexInstanceNorm()

    def forward(self, _input):
        """Forward pass of :class:`ComplexNormWrapper`."""
        # compute complex instance norm on sample of size [nBatch, nSmaps, nFE, nPE, 2]
        self.complex_instance_norm.set_normalization(_input)
        output = self.complex_instance_norm.normalize(_input)

        # re-shape data from [nBatch, nSmaps, nFE, nPE, 2] to [nBatch*nSmaps, 2, nFE, nPE]
        shp = output.shape
        output = output.view(shp[0] * shp[1], *shp[2:]).permute(0, 3, 1, 2)

        # apply denoising
        output = self.model(output)

        # re-shape data from [nBatch*nSmaps, 2, nFE, nPE]
        # to [nBatch, nSmaps, nFE, nPE, 2]
        output = output.permute(0, 2, 3, 1).view(*shp)
        # unnormalize
        output = self.complex_instance_norm.unnormalize(output)
        return output


class SensitivityNetwork(torch.nn.Module):
    """Sensitivity network with data term based on forward and adjoint containing the sensitivity maps"""

    def __init__(
        self,
        num_iter,
        model,
        datalayer,
        shared_params=True,
        save_space=False,
        reset_cache=False,
    ):
        """Init :class:`SensitivityNetwork`.

        Parameters
        ----------
        num_iter : int
            Number of iterations.
        model : torch.nn.Module
            Model to be used for the forward and adjoint.
        datalayer : torch.nn.Module
            Data layer to be used for the forward and adjoint.
        shared_params : bool, optional
            If True, the parameters of the model are shared between the forward and adjoint. Default is ``True``.
        save_space : bool, optional
            If True, the adjoint is computed in the forward pass. Default is ``False``.
        reset_cache : bool, optional
            If True, the adjoint is computed in the forward pass. Default is ``False``.
        """
        super().__init__()

        self.shared_params = shared_params
        self.num_iter = 1 if self.shared_params else num_iter
        self.num_iter_total = num_iter
        self.is_trainable = [True] * num_iter

        # setup the modules
        self.gradR = torch.nn.ModuleList([ComplexNormWrapper(model) for _ in range(self.num_iter)])
        self.gradD = torch.nn.ModuleList([datalayer for _ in range(self.num_iter)])

        self.save_space = save_space
        self.reset_cache = reset_cache

    def forward(self, x, y, smaps, mask):
        """Forward pass of :class:`SensitivityNetwork`.

        Parameters
        ----------
        x : torch.Tensor
            _input data.
        y : torch.Tensor
            Subsampled k-space data.
        smaps : torch.Tensor
            Coil sensitivity maps.
        mask : torch.Tensor
            Sampling mask.
        """
        if self.save_space:
            return self.forward_save_space(x, y, smaps, mask)

        x_all = [x]
        x_half_all = []
        if self.shared_params:
            num_iter = self.num_iter_total
        else:
            num_iter = min(np.where(self.is_trainable)[0][-1] + 1, self.num_iter)

        for i in range(num_iter):
            x_thalf = x - self.gradR[i % self.num_iter](x)
            x = self.gradD[i % self.num_iter](x_thalf, y, smaps, mask)
            x_all.append(x)
            x_half_all.append(x_thalf)

        return x_all[-1]

    def forward_save_space(self, x, y, smaps, mask):
        """Forward pass of :class:`SensitivityNetwork` with saving space.

        Parameters
        ----------
        x : torch.Tensor
            _input data.
        y : torch.Tensor
            Subsampled k-space data.
        smaps : torch.Tensor
            Coil sensitivity maps.
        mask : torch.Tensor
            Sampling mask.
        """
        if self.shared_params:
            num_iter = self.num_iter_total
        else:
            num_iter = min(np.where(self.is_trainable)[0][-1] + 1, self.num_iter)

        for i in range(num_iter):
            x_thalf = x - self.gradR[i % self.num_iter](x)
            x = self.gradD[i % self.num_iter](x_thalf, y, smaps, mask)

            # would run out of memory at test time if this is False for some cases
            if self.reset_cache:
                torch.cuda.empty_cache()
                torch.backends.cuda.cufft_plan_cache.clear()

        return x

    def freeze(self, i):
        """freeze parameter of cascade i"""
        for param in self.gradR[i].parameters():
            param.require_grad_ = False
        self.is_trainable[i] = False

    def unfreeze(self, i):
        """freeze parameter of cascade i"""
        for param in self.gradR[i].parameters():
            param.require_grad_ = True
        self.is_trainable[i] = True

    def freeze_all(self):
        """freeze parameter of cascade i"""
        for i in range(self.num_iter):
            self.freeze(i)

    def unfreeze_all(self):
        """freeze parameter of cascade i"""
        for i in range(self.num_iter):
            self.unfreeze(i)

    def copy_params(self, src_i, trg_j):
        """copy i-th cascade net parameters to j-th cascade net parameters"""
        src_params = self.gradR[src_i].parameters()
        trg_params = self.gradR[trg_j].parameters()

        for trg_param, src_param in zip(trg_params, src_params):
            trg_param.data.copy_(src_param.data)

    def stage_training_init(self):
        """set stage training flag to True"""
        self.freeze_all()
        self.unfreeze(0)
        print(self.is_trainable)

    def stage_training_transition_i(self, copy=False):
        """set stage training flag to True"""
        if self.shared_params:
            return

        # if all unlocked, don't do anything
        if not np.all(self.is_trainable):
            for i in range(self.num_iter):
                # if last cascade is reached, unlock all
                if i == self.num_iter - 1:
                    self.unfreeze_all()
                    break

                # freeze current i, unlock next. copy parameter if specified
                if self.is_trainable[i]:
                    self.freeze(i)
                    self.unfreeze(i + 1)
                    if copy:
                        self.copy_params(i, i + 1)
                    break



# coding=utf-8
__author__ = "Dimitris Karkalousos"

from typing import Sequence

import torch

from atommic.collections.common.parts.fft import fft2, ifft2


def log_likelihood_gradient(
    prediction: torch.Tensor,
    masked_kspace: torch.Tensor,
    sensitivity_maps: torch.Tensor,
    mask: torch.Tensor,
    sigma: float,
    fft_centered: bool,
    fft_normalization: str,
    spatial_dims: Sequence[int],
    coil_dim: int,
) -> torch.Tensor:
    """Computes the gradient of the log-likelihood function.

    Parameters
    ----------
    prediction : torch.Tensor
        Initial guess for the reconstruction. Shape [batch_size, height, width, 2].
    masked_kspace : torch.Tensor
        Subsampled k-space data. Shape [batch_size, coils, height, width, 2].
    sensitivity_maps : torch.Tensor
        Coil sensitivity maps. Shape [batch_size, coils, height, width, 2].
    mask : torch.Tensor
        Subsampling mask. Shape [batch_size, 1, height, width, 1].
    sigma : float
        Noise level.
    fft_centered : bool
        Whether to center the FFT.
    fft_normalization : str
        Whether to normalize the FFT.
    spatial_dims : Sequence[int]
        Spatial dimensions of the data.
    coil_dim : int
        Dimension of the coil.

    Returns
    -------
    torch.Tensor
        Gradient of the log-likelihood function. Shape [batch_size, 4, height, width]. 4 is the stacked real and
        imaginary parts of the prediction and the real and imaginary parts of the gradient.
    """
    if coil_dim == 0:
        coil_dim += 1

    prediction_real, prediction_imaginary = map(lambda x: torch.unsqueeze(x, coil_dim), prediction.chunk(2, -1))
    sensitivity_maps_real, sensitivity_maps_imaginary = sensitivity_maps.chunk(2, -1)

    re_se = prediction_real * sensitivity_maps_real - prediction_imaginary * sensitivity_maps_imaginary
    im_se = prediction_real * sensitivity_maps_imaginary + prediction_imaginary * sensitivity_maps_real
    pred = torch.cat((re_se, im_se), -1)

    pred = fft2(pred, centered=fft_centered, normalization=fft_normalization, spatial_dims=spatial_dims)

    pred = ifft2(
        mask * (pred - masked_kspace),
        centered=fft_centered,
        normalization=fft_normalization,
        spatial_dims=spatial_dims,
    )
    pred_real, pred_imag = pred.chunk(2, -1)

    if isinstance(sigma, torch.Tensor):
        sigma = sigma.item()

    sigma = max(sigma, 1.0)  # TODO: check if we need this

    re_out = torch.sum(pred_real * sensitivity_maps_real + pred_imag * sensitivity_maps_imaginary, coil_dim) / sigma
    im_out = torch.sum(pred_imag * sensitivity_maps_real - pred_real * sensitivity_maps_imaginary, coil_dim) / sigma

    prediction_real = prediction_real.squeeze(coil_dim)
    prediction_imaginary = prediction_imaginary.squeeze(coil_dim)

    return torch.cat((prediction_real, prediction_imaginary, re_out, im_out), -1).permute(0, 3, 1, 2)
